
/* Linux driver for Disk-On-Chip 2000       */
/* (c) 1999 Machine Vision Holdings, Inc.   */
/* Author: David Woodhouse <dwmw2@mvhi.com> */
/* $Id: doc2000.c,v 1.30 2000/10/13 16:18:58 mdeans Exp $ */

#include <linux/kernel.h>
#include <linux/module.h>
#include <asm/errno.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/miscdevice.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/malloc.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/types.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/nand_ids.h>
#include <linux/mtd/doc2000.h>

//#define PRERELEASE

static int doc_read (struct mtd_info *mtd, loff_t from, size_t len, size_t *retlen, u_char *buf);
static int doc_write (struct mtd_info *mtd, loff_t to, size_t len, size_t *retlen, const u_char *buf);
static int doc_read_ecc (struct mtd_info *mtd, loff_t from, size_t len, size_t *retlen, u_char *buf, u_char *eecbuf);
static int doc_write_ecc (struct mtd_info *mtd, loff_t to, size_t len, size_t *retlen, const u_char *buf, u_char *eccbuf);
static int doc_read_oob(struct mtd_info *mtd, loff_t ofs, size_t len, size_t *retlen, u_char *buf);
static int doc_write_oob(struct mtd_info *mtd, loff_t ofs, size_t len, size_t *retlen, const u_char *buf);
static int doc_erase (struct mtd_info *mtd, struct erase_info *instr);


static struct mtd_info *doc2klist = NULL;


/* Perform the required delay cycles by reading from the NOP register */
static void DoC_Delay(unsigned long docptr, unsigned short cycles)
{
	volatile char dummy;
	int i;

	for (i = 0; i < cycles; i++)
		dummy = ReadDOC(docptr, NOP);
}

/* DOC_WaitReady: Wait for RDY line to be asserted by the flash chip */
static int _DoC_WaitReady (unsigned long docptr)
{
	//long c=-1;
	short c=-1;

	DEBUG(MTD_DEBUG_LEVEL2,"_DoC_WaitReady called for out-of-line wait\n");

	/* Out-of-line routine to wait for chip response */
	while (!(ReadDOC(docptr, CDSNControl) & CDSN_CTRL_FR_B) && --c)
		;

	if (c == 0)
		DEBUG(MTD_DEBUG_LEVEL2, "_DoC_WaitReady timed out.\n");
	
	return (c==0);
}

static inline int DoC_WaitReady(unsigned long docptr) 
{
	/* This is inline, to optimise the common case, where it's ready instantly */
	volatile char dummy;
	int ret = 0;

        /* Out-of-line routine to wait for chip response */
        /* TPW: Add 4 reads - see Software Requirement 2.3.2 or 11.4.2 */
        dummy = ReadDOC(docptr, NOP);
        dummy = ReadDOC(docptr, NOP);
        dummy = ReadDOC(docptr, NOP);
        dummy = ReadDOC(docptr, NOP);
	
	if (!(ReadDOC(docptr, CDSNControl) & CDSN_CTRL_FR_B))
		ret =  _DoC_WaitReady(docptr); /* Call the out-of-line routine to wait */
	
        /* TPW: Add 2 reads - see Software Requirement 2.3.2 or 11.4.2 */
        dummy = ReadDOC(docptr, NOP);
        dummy = ReadDOC(docptr, NOP);

	return ret;
}


/* DoC_Command: Send a flash command to the flash chip */

static inline int DoC_Command(struct DiskOnChip *doc, unsigned char command, unsigned char xtraflags)
{
	unsigned long docptr;

	docptr = doc->virtadr;

	if( doc->ChipID == DOC_ChipID_Doc2k )
		xtraflags |= CDSN_CTRL_FLASH_IO;

	/* Assert the CLE (Command Latch Enable) line to the flash chip */
	WriteDOC( xtraflags | CDSN_CTRL_CLE | CDSN_CTRL_CE, docptr, CDSNControl);
	DoC_Delay(docptr, 4); /* Software requirement 11.4.3 for Millenium */
	
	if( doc->ChipID == DOC_ChipID_DocMil ) {
		WriteDOC(command, docptr, CDSNSlowIO);
	} 

	/* Send the command */
	WriteDOC_(command, docptr, doc->ioreg);
 
	/* Lower the CLE line */
	WriteDOC( xtraflags | CDSN_CTRL_CE, docptr, CDSNControl);
	DoC_Delay(docptr, 4); /* Software requirement 11.4.3 for Millenium */
	
	/* Wait for the chip to respond - Software requirement 11.4.1 (extended for any command) */
	return DoC_WaitReady(docptr);
}

/* DoC_Address: Set the current address for the flash chip */

static int DoC_Address (struct DiskOnChip *doc, int numbytes, unsigned long ofs,
			       unsigned char xtraflags1, unsigned char xtraflags2)
{
	unsigned long docptr;
	int i;

	docptr = doc->virtadr;

	if( doc->ChipID == DOC_ChipID_Doc2k ) 
		xtraflags1 |= CDSN_CTRL_FLASH_IO;

	/* Assert the ALE (Address Latch Enable line to the flash chip */
 	WriteDOC( xtraflags1 | CDSN_CTRL_ALE | CDSN_CTRL_CE,
		 docptr, CDSNControl);
	DoC_Delay(docptr, 4); /* Software requirement 11.4.3 for Millenium */

	/* Send the address */ 
	/* Devices with 256-byte page are addressed as:
		Column (bits 0-7), Page (bits 8-15, 16-23, 24-31)
		* there is no device on the market with page256
		  and more than 24 bits.
	   Devices with 512-byte page are addressed as:
		Column (bits 0-7), Page (bits 9-16, 17-24, 25-31)
		* 25-31 is sent only if the chip support it.
		* bit 8 changes the read command to be sent
		  (NAND_CMD_READ0 or NAND_CMD_READ1).
	*/

	if (numbytes == ADDR_COLUMN || numbytes == ADDR_COLUMN_PAGE ) {
		if( doc->ChipID == DOC_ChipID_DocMil )
			WriteDOC(ofs & 0xff, docptr, CDSNSlowIO);
		WriteDOC_(ofs & 0xff, docptr, doc->ioreg);
	}
	
	if ( doc->page256 ) {
		ofs = ofs >> 8;
	} else {
		ofs = ofs >> 9;
	}

	if (numbytes == ADDR_PAGE || numbytes == ADDR_COLUMN_PAGE ) {	
		for( i = 0; i < doc->pageadrlen; i++, ofs = ofs >> 8 ) {
			if( doc->ChipID == DOC_ChipID_DocMil )
				WriteDOC(ofs & 0xff, docptr, CDSNSlowIO);
			WriteDOC_(ofs & 0xff, docptr, doc->ioreg);
		}
	}

	DoC_Delay(docptr, 2); /* Needed for some slow flash chips. mf. */

	/* FIXME: The SlowIO's for millenium could be replaced by 
	   a single WritePipeTerm here. mf. */

	/* Lower the ALE line */
	WriteDOC( xtraflags1 | xtraflags2 | CDSN_CTRL_CE, docptr, CDSNControl);
	DoC_Delay(docptr, 4); /* Software requirement 11.4.3 for Millenium */
	
	/* Wait for the chip to respond - Software requirement 11.4.1 */
	return DoC_WaitReady(docptr);
}

/* Read a buffer from DoC, taking care of Millenium odditys */
static void DoC_ReadBuf(struct DiskOnChip *doc, u_char *buf, int len )
{
	int dummy;
	int modulus = 0xffff;
	unsigned long docptr;
	int i;

	docptr = doc->virtadr;

	if( len <= 0 )
		return;

	if( doc->ChipID == DOC_ChipID_DocMil ) {
		/* Read the data via the internal pipeline through CDSN IO register,
			see Pipelined Read Operations 11.3 */
		dummy = ReadDOC(docptr, ReadPipeInit);

		/* Millenium should use the LastDataRead register - Pipeline Reads */
		len--;

		/* This is needed for correctly ECC calculation */
		modulus = 0xff;
	}

	for (i=0; i < len ; i++)
		buf[i] = ReadDOC_(docptr, doc->ioreg + (i & modulus) );

	if( doc->ChipID == DOC_ChipID_DocMil ) {
		buf[i] = ReadDOC(docptr, LastDataRead);
	}
}

/* Write a buffer to DoC, taking care of Millenium odditys */
static void DoC_WriteBuf(struct DiskOnChip *doc, const u_char *buf, int len )
{
	unsigned long docptr;
	int i;

	docptr = doc->virtadr;

	if( len <= 0 )
		return;

	for (i=0; i < len ; i++)
		WriteDOC_(buf[i], docptr, doc->ioreg + i);

	if( doc->ChipID == DOC_ChipID_DocMil ) {
		WriteDOC(0x00, docptr, WritePipeTerm);
	}
}


/* DoC_SelectChip: Select a given flash chip within the current floor */

static inline int DoC_SelectChip(unsigned long docptr, int chip)
{
	/* Software requirement 11.4.4 before writing DeviceSelect */
	/* Deassert the CE line to eliminate glitches on the FCE# outputs */
	WriteDOC( CDSN_CTRL_WP, docptr, CDSNControl);
	DoC_Delay(docptr, 4); /* Software requirement 11.4.3 for Millenium */

	/* Select the individual flash chip requested */
	WriteDOC( chip, docptr, CDSNDeviceSelect);
	DoC_Delay(docptr, 4);

	/* Reassert the CE line */
	WriteDOC( CDSN_CTRL_CE | CDSN_CTRL_FLASH_IO | CDSN_CTRL_WP, docptr, CDSNControl);
	DoC_Delay(docptr, 4); /* Software requirement 11.4.3 for Millenium */
	
	/* Wait for it to be ready */
	return DoC_WaitReady(docptr);
}

/* DoC_SelectFloor: Select a given floor (bank of flash chips) */

static inline int DoC_SelectFloor(unsigned long docptr, int floor)
{
	/* Select the floor (bank) of chips required */
	WriteDOC( floor, docptr, FloorSelect);

	/* Wait for the chip to be ready */
	return DoC_WaitReady(docptr);
}
  
/* DoC_IdentChip: Identify a given NAND chip given {floor,chip} */

static int DoC_IdentChip(struct DiskOnChip *doc, int floor, int chip)
{
	int mfr, id, i;
	volatile char dummy;

	/* Page in the required floor/chip */
	DoC_SelectFloor(doc->virtadr, floor);
	DoC_SelectChip(doc->virtadr, chip);

	/* Reset the chip */
	if (DoC_Command(doc, NAND_CMD_RESET, CDSN_CTRL_WP)) {
		DEBUG(MTD_DEBUG_LEVEL2, "DoC_Command (reset) for %d,%d returned true\n", floor,chip);
		return 0;
	}

  
	/* Read the NAND chip ID: 1. Send ReadID command */ 
	if(DoC_Command(doc, NAND_CMD_READID, CDSN_CTRL_WP)) {
		DEBUG(MTD_DEBUG_LEVEL2,"DoC_Command (ReadID) for %d,%d returned true\n", floor,chip);
		return 0;
	}

	/* Read the NAND chip ID: 2. Send address byte zero 
	 */ 
	DoC_Address(doc, ADDR_COLUMN, 0, CDSN_CTRL_WP, 0);
 

	if( doc->ChipID == DOC_ChipID_DocMil ) {
		/* CDSN Slow IO register see Software Requirement 11.4 item 5.*/
		dummy = ReadDOC(doc->virtadr, CDSNSlowIO);
		DoC_Delay(doc->virtadr, 2);
	}

	/* Read the manufacturer and device id codes from the device */
	mfr = ReadDOC_(doc->virtadr, doc->ioreg);

	if( doc->ChipID == DOC_ChipID_DocMil ) {
		/* CDSN Slow IO register see Software Requirement 11.4 item 5.*/
		dummy = ReadDOC(doc->virtadr, CDSNSlowIO);
		DoC_Delay(doc->virtadr, 2);
	}

	id = ReadDOC_(doc->virtadr, doc->ioreg);
	
	/* No response - return failure */
	if (mfr == 0xff || mfr == 0)
		return 0;
	
	/* Check it's the same as the first chip we identified. 
	 * M-Systems say that any given DiskOnChip device should only
	 * contain _one_ type of flash part, although that's not a 
	 * hardware restriction. */
	if (doc->mfr) {
		if (doc->mfr == mfr && doc->id == id)
			return 1; /* This is another the same the first */
		else
			printk(KERN_WARNING "Flash chip at floor %d, chip %d is different:\n",
			       floor, chip);
	}
	
	/* Print and store the manufacturer and ID codes. */
	for (i = 0; nand_flash_ids[i].name != NULL; i++) {
		if (mfr == nand_flash_ids[i].manufacture_id &&
		    id == nand_flash_ids[i].model_id) {
			printk(KERN_INFO "Flash chip found: Manufacture ID: %2.2X, "
			       "Chip ID: %2.2X (%s)\n",
			       mfr, id, nand_flash_ids[i].name);
			if (!doc->mfr) {
				doc->mfr = mfr;
				doc->id = id;
				doc->chipshift = nand_flash_ids[i].chipshift;
				doc->page256 = nand_flash_ids[i].page256;
				doc->pageadrlen = nand_flash_ids[i].pageadrlen;
				doc->erasesize = nand_flash_ids[i].erasesize;
				return 1;
			}
			return 0;
		}
	}
  
	
	/* We haven't fully identified the chip. Print as much as we know. */
	printk(KERN_WARNING "Unknown flash chip found: %2.2X %2.2X\n", id, mfr);
	
	printk(KERN_WARNING "Please report to David.Woodhouse@mvhi.com\n");
	return 0;
}     

/* DoC_ScanChips: Find all NAND chips present in a DiskOnChip, and identify them */

static void DoC_ScanChips(struct DiskOnChip *this)
{
	int floor, chip;
	int numchips[MAX_FLOORS];
	int ret = 1;
	
	this->numchips = 0;
	this->mfr = 0;
	this->id = 0;
	
	/* For each floor, find the number of valid chips it contains */
	for (floor = 0 ; floor < MAX_FLOORS ; floor++) {
		ret = 1;
		numchips[floor]=0;
		for (chip = 0 ; chip < MAX_CHIPS && ret != 0; chip++ ) {
			
			ret = DoC_IdentChip(this, floor, chip);
			if (ret) {
				numchips[floor]++;
				this->numchips++;
			}
		}
	}
	
	/* If there are none at all that we recognise, bail */
	if (!this->numchips) {
		printk("No flash chips recognised.\n");
		return;
	}
	
	/* Allocate an array to hold the information for each chip */
	this->chips = kmalloc(sizeof(struct Nand) * this->numchips, GFP_KERNEL);
	if (!this->chips){
		printk("No memory for allocating chip info structures\n");
		return;
	}
	
	ret = 0;
	
	/* Fill out the chip array with {floor, chipno} for each 
	 * detected chip in the device. */
	for (floor = 0; floor < MAX_FLOORS; floor++) {
		for (chip = 0 ; chip < numchips[floor] ; chip++) {
			this->chips[ret].floor = floor;
			this->chips[ret].chip = chip;
			this->chips[ret].curadr = 0;
			this->chips[ret].curmode = 0x50;
			ret++;
		}
	}

	/* Calculate and print the total size of the device */
	this->totlen = this->numchips * (1 << this->chipshift);

	printk(KERN_INFO "%d flash chips found. Total DiskOnChip size: %ld Mb\n", this->numchips ,
	       this->totlen >> 20);
}


static int DoC2k_is_alias(struct DiskOnChip *doc1, struct DiskOnChip *doc2)
{
	int tmp1, tmp2, retval;
	if (doc1->physadr == doc2->physadr)
		return 1;

	/* Use the alias resolution register which was set aside for this
	 * purpose. If it's value is the same on both chips, they might
	 * be the same chip, and we write to one and check for a change in
	 * the other. It's unclear if this register is usuable in the
	 * DoC 2000 (it's in the Millenium docs), but it seems to work. */
	tmp1 = ReadDOC(doc1->virtadr, AliasResolution);
	tmp2 = ReadDOC(doc2->virtadr, AliasResolution);
	if (tmp1 != tmp2)
		return 0;
	
	WriteDOC((tmp1+1) % 0xff, doc1->virtadr, AliasResolution);
	tmp2 = ReadDOC(doc2->virtadr, AliasResolution);
	if (tmp2 == (tmp1+1) % 0xff)
		retval = 1;
	else
		retval = 0;

	/* Restore register contents.  May not be necessary, but do it just to
	 * be safe. */
	WriteDOC(tmp1, doc1->virtadr, AliasResolution);
	
	return retval;
}


void DoC2k_init(struct mtd_info *mtd)
{
	struct DiskOnChip *this = (struct DiskOnChip *)mtd->priv;
	struct DiskOnChip *old = NULL;

	/* We must avoid being called twice for the same device. */

	if (doc2klist)
		old = (struct DiskOnChip *)doc2klist->priv;

	while (old) {
		if (DoC2k_is_alias(old, this)) {
			printk(KERN_NOTICE "Ignoring DiskOnChip 2000 at 0x%lX - already configured\n",
			       this->physadr);
			iounmap((void *)this->virtadr);
			kfree(mtd);
			return;
		}
		if (old->nextdoc)
			old = (struct DiskOnChip *)old->nextdoc->priv;
		else
			old = NULL;
	}
			
	
	switch(this->ChipID) {
	case DOC_ChipID_Doc2k:
		mtd->name = "DiskOnChip 2000";
		this->ioreg=DoC_2k_CDSN_IO;
		break;
	case DOC_ChipID_DocMil:
		mtd->name = "DiskOnChip Millenium";
		this->ioreg=DoC_Mil_CDSN_IO;
		break;
	}

	printk(KERN_NOTICE "%s found at address 0x%lX\n", mtd->name, this->physadr);

	mtd->type = MTD_NANDFLASH;
	mtd->flags = MTD_CAP_NANDFLASH;
	mtd->size = 0;
	mtd->erasesize = 0;
	mtd->oobblock = 512;
	mtd->oobsize = 16;
	mtd->module = THIS_MODULE;
	mtd->erase = doc_erase;
	mtd->point = NULL;
	mtd->unpoint = NULL;
	mtd->read = doc_read;
	mtd->write = doc_write;
	mtd->read_ecc = doc_read_ecc;
	mtd->write_ecc = doc_write_ecc;
	mtd->read_oob = doc_read_oob;
	mtd->write_oob = doc_write_oob;
	mtd->sync = NULL;

	this->totlen = 0;
	this->numchips = 0;
	
	this->curfloor = -1;
	this->curchip = -1;
	
	/* Ident all the chips present. */
	DoC_ScanChips(this);
	
	if (!this->totlen) {
		kfree(mtd);
		iounmap((void *)this->virtadr);
	} else {
		this->nextdoc = doc2klist;
		doc2klist = mtd;
		mtd->size  = this->totlen;
		mtd->erasesize = this->erasesize;
		add_mtd_device(mtd);
		return;
	}
}


EXPORT_SYMBOL_NOVERS(DoC2k_init);

static int doc_read (struct mtd_info *mtd, loff_t from, size_t len, size_t *retlen, u_char *buf)
{
	/* Just a special case of doc_read_ecc */
	return doc_read_ecc(mtd, from, len, retlen, buf, NULL);
}

static int doc_read_ecc (struct mtd_info *mtd, loff_t from, size_t len, size_t *retlen, u_char *buf, u_char *eccbuf)
{
	struct DiskOnChip *this = (struct DiskOnChip *)mtd->priv;
	unsigned long docptr;
	struct Nand *mychip;
	unsigned char syndrome[6];
	volatile char dummy;
	int i, len256 = 0;

	docptr = this->virtadr;

	/* Don't allow read past end of device */
	if (from >= this->totlen)
		return -EINVAL;
	
	/* Don't allow a single read to cross a 512-byte block boundary */
	if (from + len > ( (from | 0x1ff) + 1)) 
		len = ((from | 0x1ff) + 1) - from;

	/* The ECC will not be calculated correctly if less than 512 is read */
	if ( len != 0x200 && eccbuf )
		printk(KERN_WARNING "ECC needs a full sector read (adr: %lx size %lx)\n", (long) from, (long) len);

	/* printk("DoC_Read (adr: %lx size %lx)\n", (long) from, (long) len); */


	/* Find the chip which is to be used and select it */
	mychip = &this->chips[from >> (this->chipshift)];
	
	if (this->curfloor != mychip->floor) {
		DoC_SelectFloor(docptr, mychip->floor);
		DoC_SelectChip(docptr, mychip->chip);
	}
	else if (this->curchip != mychip->chip) {
		DoC_SelectChip(docptr, mychip->chip);
	}
	
	this->curfloor = mychip->floor;
	this->curchip = mychip->chip;

	DoC_Command(this, (!this->page256 && (from & 0x100)) ? NAND_CMD_READ1 : NAND_CMD_READ0,
		    CDSN_CTRL_WP);
	DoC_Address(this, ADDR_COLUMN_PAGE, from, CDSN_CTRL_WP, CDSN_CTRL_ECC_IO);

	if (eccbuf) {
		/* Prime the ECC engine */
		WriteDOC ( DOC_ECC_RESET, docptr, ECCConf);
		WriteDOC ( DOC_ECC_EN, docptr, ECCConf);
	} else {
		/* disable the ECC engine */
		WriteDOC (DOC_ECC_RESET, docptr, ECCConf);
		WriteDOC (DOC_ECC_DIS, docptr, ECCConf);
	}

	/* treat crossing 256-byte sector for 2M x 8bits devices */
	if ( this->page256 && from+len > (from | 0xff) + 1 ) {
		len256 = (from | 0xff) + 1 - from;
		DoC_ReadBuf(this, buf, len256 );

		DoC_Command(this, NAND_CMD_READ0, CDSN_CTRL_WP);
		DoC_Address(this, ADDR_COLUMN_PAGE, from + len256,
			CDSN_CTRL_WP, CDSN_CTRL_ECC_IO);
	}

	DoC_ReadBuf(this, &buf[len256], len-len256);

	/* Let the caller know we completed it */
	*retlen = len;

	if (eccbuf) {
		/* Read the ECC data through the DiskOnChip ECC logic */
		/* Note: this will work even with 2M x 8bit devices as   */
		/*       they have 8 bytes of OOB per 256 page. mf.      */
		DoC_ReadBuf(this, eccbuf, 6 );
		
		/* Flush the pipeline */
		if( this->ChipID == DOC_ChipID_DocMil ) {
			dummy = ReadDOC(docptr, ECCConf);
			dummy = ReadDOC(docptr, ECCConf);
			i = ReadDOC(docptr, ECCConf);
		} else {
			dummy = ReadDOC(docptr, 2k_ECCStatus);
			dummy = ReadDOC(docptr, 2k_ECCStatus);
			i = ReadDOC(docptr, 2k_ECCStatus);
		}
		
		/* Check the ECC Status */
		if (i & 0x80) {
			/* There was an ECC error */
			printk("DiskOnChip ECC Error: Read at %lx\n", (long)from);

			/* FIXME: Implement ECC error correction, don't just whinge */

			/* Read the ECC syndrom through the DiskOnChip ECC logic.
			   These syndrome will be all ZERO when there is no error */
			for (i = 0; i < 6; i++) {
				syndrome[i] = ReadDOC(docptr, ECCSyndrome0 + i);
			}

			printk("ECC syndrom: %2.2X %2.2X %2.2X %2.2X %2.2X %2.2X\n",
			       syndrome[0], syndrome[1], syndrome[2],
			       syndrome[3], syndrome[4], syndrome[5]);

			/* We return error, but have actually done the read. Not that
			   this can be told to user-space, via sys_read(), but at least
			   MTD-aware stuff can know about it by checking *retlen */
			return -EIO;
		}
#ifdef PSYCHO_DEBUG
		else
			printk("ECC OK at %lx, OOB: %2.2X %2.2X %2.2X %2.2X %2.2X %2.2X\n",
			       (long)from, eccbuf[0], eccbuf[1], eccbuf[2], eccbuf[3], eccbuf[4],
			       eccbuf[5]);
#endif
		
		/* Reset the ECC engine */
		WriteDOC(DOC_ECC_RESV, docptr , ECCConf);
		
	}

	return 0;
}

static int doc_write (struct mtd_info *mtd, loff_t to, size_t len, size_t *retlen, const u_char *buf)
{
	return doc_write_ecc(mtd, to, len, retlen, buf, NULL);
}

static int doc_write_ecc (struct mtd_info *mtd, loff_t to, size_t len, size_t *retlen, const u_char *buf, u_char *eccbuf)
{
	struct DiskOnChip *this = (struct DiskOnChip *)mtd->priv;
	int di; /* Yes, DI is a hangover from when I was disassembling the binary driver */
	unsigned long docptr;
	volatile char dummy;
	int len256 = 0;
	struct Nand *mychip;

	docptr = this->virtadr;

	/* Don't allow write past end of device */
	if (to >= this->totlen)
		return -EINVAL;

	/* Don't allow a single write to cross a 512-byte block boundary */
	if (to + len > ( (to | 0x1ff) + 1)) 
		len = ((to | 0x1ff) + 1) - to;

	/* The ECC will not be calculated correctly if less than 512 is written */
	if ( len != 0x200 && eccbuf )
		printk(KERN_WARNING "ECC needs a full sector write (adr: %lx size %lx)\n", (long) to, (long) len);

	/* printk("DoC_Write (adr: %lx size %lx)\n", (long) to, (long) len); */

	/* Find the chip which is to be used and select it */
	mychip = &this->chips[to >> (this->chipshift)];
	
	if (this->curfloor != mychip->floor) {
		DoC_SelectFloor(docptr, mychip->floor);
		DoC_SelectChip(docptr, mychip->chip);
	}
	else if (this->curchip != mychip->chip) {
		DoC_SelectChip(docptr, mychip->chip);
	}
	
	this->curfloor = mychip->floor;
	this->curchip = mychip->chip;
	
	/* Set device to main plane of flash */
	DoC_Command(this, NAND_CMD_RESET, CDSN_CTRL_WP);
	DoC_Command(this, (!this->page256 && (to & 0x100)) ? NAND_CMD_READ1 : NAND_CMD_READ0,
			 CDSN_CTRL_WP);

	DoC_Command(this, NAND_CMD_SEQIN, 0);
	DoC_Address(this, ADDR_COLUMN_PAGE, to, 0, CDSN_CTRL_ECC_IO);

	if (eccbuf) {
		/* Prime the ECC engine */
		WriteDOC ( DOC_ECC_RESET, docptr, ECCConf);
		WriteDOC ( DOC_ECC_EN | DOC_ECC_RW, docptr, ECCConf);
	} else {
		/* disable the ECC engine */
		WriteDOC (DOC_ECC_RESET, docptr, ECCConf);
		WriteDOC (DOC_ECC_DIS, docptr, ECCConf);
	}

	/* treat crossing 256-byte sector for 2M x 8bits devices */
	if ( this->page256 && to+len > (to | 0xff) + 1 ) {
		len256 = (to | 0xff) + 1 - to;
		DoC_WriteBuf(this, buf, len256 );
		
		DoC_Command(this, NAND_CMD_PAGEPROG, 0);

		DoC_Command(this, NAND_CMD_STATUS, CDSN_CTRL_WP);
		/* There's an implicit DoC_WaitReady() in DoC_Command */

		if( this->ChipID == DOC_ChipID_DocMil ) {
			dummy = ReadDOC(docptr, CDSNSlowIO);
			DoC_Delay(docptr, 2);
		}

		if (ReadDOC_(docptr, this->ioreg) & 1) {
			printk("Error programming flash\n");
			/* Error in programming */
			*retlen = 0;
			return -EIO;
		}

		DoC_Command(this, NAND_CMD_SEQIN, 0);
		DoC_Address(this, ADDR_COLUMN_PAGE, to + len256, 0, CDSN_CTRL_ECC_IO);
	}

	DoC_WriteBuf(this, &buf[len256], len-len256);

	if (eccbuf) {
		WriteDOC( CDSN_CTRL_ECC_IO | CDSN_CTRL_CE , docptr, CDSNControl );
		
		if( this->ChipID == DOC_ChipID_DocMil ) {
			WriteDOC(0, docptr, NOP);
			WriteDOC(0, docptr, NOP);
			WriteDOC(0, docptr, NOP);
		} else {
			WriteDOC_( 0, docptr, this->ioreg);
			WriteDOC_( 0, docptr, this->ioreg);
			WriteDOC_( 0, docptr, this->ioreg);
		}

		/* Read the ECC data through the DiskOnChip ECC logic */
		for (di=0; di<6; di++) {
			eccbuf[di] = ReadDOC(docptr, ECCSyndrome0 + di);
		}
#ifdef PSYCHO_DEBUG
		printk("OOB data at %lx is %2.2X %2.2X %2.2X %2.2X %2.2X %2.2X\n",
		       (long) to, eccbuf[0], eccbuf[1], eccbuf[2],
		       eccbuf[3], eccbuf[4], eccbuf[5]       );
#endif
		/* Reset the ECC engine */
		WriteDOC(DOC_ECC_RESV, docptr , ECCConf);
		
	}

	DoC_Command(this, NAND_CMD_PAGEPROG, 0);

	DoC_Command(this, NAND_CMD_STATUS, CDSN_CTRL_WP);
	/* There's an implicit DoC_WaitReady() in DoC_Command */

	if( this->ChipID == DOC_ChipID_DocMil ) {
		dummy = ReadDOC(docptr, CDSNSlowIO);
		DoC_Delay(docptr, 2);
	}

	if (ReadDOC_(docptr, this->ioreg) & 1) {
		printk("Error programming flash\n");
		/* Error in programming */
		*retlen = 0;
		return -EIO;
	}

	/* Let the caller know we completed it */
	*retlen = len;

	return 0;
}

static int doc_read_oob(struct mtd_info *mtd, loff_t ofs, size_t len, size_t *retlen, u_char *buf)
{
	struct DiskOnChip *this = (struct DiskOnChip *)mtd->priv;
	int len256 = 0;
	unsigned long docptr;
	struct Nand *mychip;
	
	docptr = this->virtadr;
	
	mychip = &this->chips[ofs >> this->chipshift];
	
	if (this->curfloor != mychip->floor) {
		DoC_SelectFloor(docptr, mychip->floor);
		DoC_SelectChip(docptr, mychip->chip);
	}
	else if (this->curchip != mychip->chip) {
		DoC_SelectChip(docptr, mychip->chip);
	}
	this->curfloor = mychip->floor;
	this->curchip = mychip->chip;
	
	/* update address for 2M x 8bit devices. OOB starts on the second */
	/* page to maintain compatibility with doc_read_ecc. */
	if( this->page256 ) {
		if( !(ofs & 0x8) )
			ofs += 0x100;
		else
			ofs -= 0x8;
	} 
	
	DoC_Command(this, NAND_CMD_READOOB, CDSN_CTRL_WP);
	DoC_Address(this, ADDR_COLUMN_PAGE, ofs, CDSN_CTRL_WP, 0);
	
	/* treat crossing 8-byte OOB data for 2M x 8bit devices */
	/* Note: datasheet says it should automaticaly wrap to the */
	/*       next OOB block, but it didn't work here. mf.      */
	if ( this->page256 && ofs+len > (ofs | 0x7) + 1 ) {
		len256 = (ofs | 0x7) + 1 - ofs;
		DoC_ReadBuf(this, buf, len256 );
	
		DoC_Command(this, NAND_CMD_READOOB, CDSN_CTRL_WP);
		DoC_Address(this, ADDR_COLUMN_PAGE, ofs & (~0x1ff), CDSN_CTRL_WP, 0);
	}
	
	DoC_ReadBuf(this, &buf[len256], len - len256 );
	
	*retlen = len;
	return 0;

}

static int doc_write_oob(struct mtd_info *mtd, loff_t ofs, size_t len, size_t *retlen, const u_char *buf)
{
	struct DiskOnChip *this = (struct DiskOnChip *)mtd->priv;
	int len256 = 0;
	unsigned long docptr;
	struct Nand *mychip;
	int dummy;

	//	printk("doc_write_oob(%lx, %d): %2.2X %2.2X %2.2X %2.2X ... %2.2X %2.2X .. %2.2X %2.2X\n",(long)ofs, len,
	//   buf[0], buf[1], buf[2], buf[3], buf[8], buf[9], buf[14],buf[15]);

	docptr = this->virtadr;
	
	mychip = &this->chips[ofs >> this->chipshift];
	
	if (this->curfloor != mychip->floor) {
		DoC_SelectFloor(docptr, mychip->floor);
		DoC_SelectChip(docptr, mychip->chip);
	}
	else if (this->curchip != mychip->chip) {
		DoC_SelectChip(docptr, mychip->chip);
	}
	this->curfloor = mychip->floor;
	this->curchip = mychip->chip;
	
	DoC_Command(this, NAND_CMD_RESET, CDSN_CTRL_WP);
	DoC_Command(this, NAND_CMD_READOOB, CDSN_CTRL_WP);

	/* update address for 2M x 8bit devices. OOB starts on the second */
	/* page to maintain compatibility with doc_read_ecc. */
	if( this->page256 ) {
		if( !(ofs & 0x8) )
			ofs += 0x100;
		else
			ofs -= 0x8;
	} 

	DoC_Command(this, NAND_CMD_SEQIN, 0);
	DoC_Address(this, ADDR_COLUMN_PAGE, ofs, 0, 0);
	
	/* treat crossing 8-byte OOB data for 2M x 8bit devices */
	/* Note: datasheet says it should automaticaly wrap to the */
	/*       next OOB block, but it didn't work here. mf.      */
	if ( this->page256 && ofs+len > (ofs | 0x7) + 1 ) {
		len256 = (ofs | 0x7) + 1 - ofs;
		DoC_WriteBuf(this, buf, len256 );

		DoC_Command(this, NAND_CMD_PAGEPROG, 0);
		DoC_Command(this, NAND_CMD_STATUS, 0);
		/* DoC_WaitReady() is implicit in DoC_Command */

		if( this->ChipID == DOC_ChipID_DocMil ) {
			dummy = ReadDOC(docptr, CDSNSlowIO);
			DoC_Delay(docptr, 2);
		}

		if (ReadDOC_(docptr, this->ioreg) & 1) {
			printk("Error programming oob data\n");
			/* There was an error */
			*retlen = 0;
			return -EIO;
		}
		DoC_Command(this, NAND_CMD_SEQIN, 0);
		DoC_Address(this, ADDR_COLUMN_PAGE, ofs & (~0x1ff), 0, 0);
	}
	
	DoC_WriteBuf(this, &buf[len256], len-len256 );

	DoC_Command(this, NAND_CMD_PAGEPROG, 0);
	DoC_Command(this, NAND_CMD_STATUS, 0);
	/* DoC_WaitReady() is implicit in DoC_Command */

	if( this->ChipID == DOC_ChipID_DocMil ) {
		dummy = ReadDOC(docptr, CDSNSlowIO);
		DoC_Delay(docptr, 2);
	}

	if (ReadDOC_(docptr, this->ioreg) & 1) {
		printk("Error programming oob data\n");
		/* There was an error */
		*retlen = 0;
		return -EIO;
	}

	*retlen = len;
	return 0;

}


int doc_erase (struct mtd_info *mtd, struct erase_info *instr)
{
	struct DiskOnChip *this = (struct DiskOnChip *)mtd->priv;
	unsigned long ofs = instr->addr;
	unsigned long len = instr->len;
	unsigned long docptr;
	struct Nand *mychip;
	
	if(len != mtd->erasesize) 
		printk(KERN_WARNING "Erase not right size (%lx != %lx)n", len, mtd->erasesize);
		
	docptr = this->virtadr;
	
	mychip = &this->chips[ofs >> this->chipshift];
	
	if (this->curfloor != mychip->floor) {
		DoC_SelectFloor(docptr, mychip->floor);
		DoC_SelectChip(docptr, mychip->chip);
	}
	else if (this->curchip != mychip->chip) {
		DoC_SelectChip(docptr, mychip->chip);
	}
	this->curfloor = mychip->floor;
	this->curchip = mychip->chip;
	
	instr->state = MTD_ERASE_PENDING;

	DoC_Command(this, NAND_CMD_ERASE1, 0);
	DoC_Address(this, ADDR_PAGE, ofs, 0, 0);
	DoC_Command(this, NAND_CMD_ERASE2, 0);

	instr->state = MTD_ERASING;

	DoC_Command(this, NAND_CMD_STATUS, CDSN_CTRL_WP);

	if (ReadDOC_(docptr, this->ioreg) & 1) {
		printk("Error writing\n");
		/* There was an error */
		instr->state = MTD_ERASE_FAILED;
	}
	else
		instr->state = MTD_ERASE_DONE;

	if (instr->callback) 
		instr->callback(instr);
			
	return 0;
}





/****************************************************************************
 *
 * Module stuff
 *
 ****************************************************************************/

#if LINUX_VERSION_CODE < 0x20300
#ifdef MODULE
#define cleanup_doc2000 cleanup_module
#endif
#define __exit
#endif


static void __exit cleanup_doc2000(void)
{
	struct mtd_info *mtd;
	struct DiskOnChip *this;

	while((mtd=doc2klist)) {
		this = (struct DiskOnChip *)mtd->priv;
		doc2klist = this->nextdoc;
			
		del_mtd_device(mtd);
			
		iounmap((void *)this->virtadr);
		kfree(this->chips);
		kfree(mtd);
	}

}

#if LINUX_VERSION_CODE > 0x20300
module_exit(cleanup_doc2000);
#endif

