/* 
 * NFTL mount code with extensive checks
 *
 * Author: Fabrice Bellard (fabrice.bellard@netgem.com) 
 * Copyright (C) 2000 Netgem S.A.
 *
 * $Id: nftlmount.c,v 1.1 2000/11/03 12:43:43 dwmw2 Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <asm/errno.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/miscdevice.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/malloc.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nftl.h>
#include <linux/mtd/compatmac.h>

#define SECTORSIZE 512

/* Find the block which contains the bad block table & various device
   informations. Also search for its copy. Update the ReplUnitTable[]
   table to indicate the bad blocks */
static int find_boot_record(struct NFTLrecord *s)
{
    struct nftl_uci1 h1;
    struct nftl_oob oob;
    unsigned int block, boot_record_count, erase_mark, i;
    int retlen;
    u8 buf[SECTORSIZE];
    struct NFTLMediaHeader *mh = &s->MediaHdr;

    s->MediaUnit = BLOCK_NIL;
    s->SpareMediaUnit = BLOCK_NIL;

    boot_record_count = 0;
    /* search for a valid boot record */
    for(block=0;block<s->nb_blocks;block++) {
        /* read ANAND header. To be safer with BIOS, also use erase
           mark as discriminant */
        if (MTD_READOOB(s->mtd, block * s->EraseSize + SECTORSIZE + 8, 8,
                        &retlen, (char *)&h1) < 0)
            continue;
        erase_mark = le16_to_cpu ((h1.EraseMark | h1.EraseMark1));
        if (erase_mark != ERASE_MARK) 
            continue;
        if (MTD_READECC(s->mtd, block * s->EraseSize, SECTORSIZE,
                        &retlen, buf, (char *)&oob) < 0)
            continue;
        memcpy(mh, buf, sizeof(struct NFTLMediaHeader));
        if (memcmp(mh->DataOrgID, "ANAND", 6) == 0) {
            /* first boot record */
            if (boot_record_count == 0) {
                /* header found : read the bad block table data */
                if (mh->UnitSizeFactor != 0xff) {
                    printk("Sorry, we don't support UnitSizeFactor of != 1 yet\n");
                    goto ReplUnitTable;
                }
                s->nb_boot_blocks = le16_to_cpu(mh->FirstPhysicalEUN);
                if ((s->nb_boot_blocks + 2) >= s->nb_blocks)
                    goto ReplUnitTable; /* small consistency check */
                s->numvunits = le32_to_cpu(mh->FormattedSize) / s->EraseSize;
                if (s->numvunits > (s->nb_blocks - s->nb_boot_blocks - 2))
                    goto ReplUnitTable; /* small consistency check */
                s->nr_sects =  s->numvunits  * (s->EraseSize / SECTORSIZE);
                s->MediaUnit = block;
                /* read the bad sector table and modify ReplUnitTable[] accordingly */
                for(i=0;i<s->nb_blocks;i++) {
                    if ((i & (SECTORSIZE - 1)) == 0) {
                        /* read one sector every SECTORSIZE blocks */
                        if (MTD_READECC(s->mtd, block * s->EraseSize + i + SECTORSIZE, SECTORSIZE,
                                        &retlen, buf, (char *)&oob) < 0)
                            goto ReplUnitTable;
                    }
                    /* no need to test bit error here because the
                       sector is corrected by ECC */
                    if (buf[i & (SECTORSIZE - 1)] != 0xff)
                        s->ReplUnitTable[i] = BLOCK_RESERVED;
                }
                boot_record_count++;
            } else if (boot_record_count == 1) {
                s->SpareMediaUnit = block;
                boot_record_count++;
                break;
            }
        }
    ReplUnitTable:
    }
    if (boot_record_count == 0) {
        /* no boot record found */
        return -1;
    } else {
        return 0;
    }
}

static int memcmpb(void *a, int c, int n)
{
    int i;
    for(i=0;i<n;i++) {
        if (c != ((unsigned char *)a)[i])
            return 1;
    }
    return 0;
}

static int check_free_sectors(struct NFTLrecord *s, unsigned int address, int len, 
                              int check_oob)
{
    int i, retlen;
    u8 buf[SECTORSIZE];

    for(i=0;i<len;i+=SECTORSIZE) {
        /* we want to read the sector without ECC check here */
        if (MTD_READECC(s->mtd, address, SECTORSIZE, &retlen, buf, NULL) < 0)
            return -1;
        if (memcmpb(buf, 0xff, SECTORSIZE) != 0)
            return -1;
        if (check_oob) {
            if (MTD_READOOB(s->mtd, address, s->mtd->oobsize, &retlen, buf) < 0)
                return -1;
            if (memcmpb(buf, 0xff, s->mtd->oobsize) != 0)
                return -1;
        }
        address += SECTORSIZE;
    }
    return 0;
}


/* format a block, check it, and tries another time if error. Return
   -1 if error */
int NFTL_formatblock(struct NFTLrecord *s, int block)
{
    int retry, retlen;
    unsigned int nb_erases, erase_mark;
    struct nftl_uci1 h1;
    struct erase_info *instr = &s->instr;

    if (MTD_READOOB(s->mtd, block * s->EraseSize + SECTORSIZE + 8, 8,
                    &retlen, (char *)&h1) < 0)
        goto default_uci1;
    erase_mark = le16_to_cpu ((h1.EraseMark | h1.EraseMark1));
    if (erase_mark != ERASE_MARK) {
    default_uci1:
        h1.EraseMark = cpu_to_le16(ERASE_MARK);
        h1.EraseMark1 = cpu_to_le16(ERASE_MARK);
        h1.WearInfo = cpu_to_le32(0);
    }

    for(retry = 0; retry < 2; retry++) {
        memset(instr, 0, sizeof(struct erase_info));
        /* XXX: use async erase interface, XXX: test return code */
        instr->addr = block * s->EraseSize;
        instr->len = s->EraseSize;
        MTD_ERASE(s->mtd, instr);

        /* XXX: is this check really necessary ? */
        if (check_free_sectors(s, instr->addr, s->EraseSize, 1) != 0)
            continue;
        break;
    }
    if (retry == 2) {
        /* could not format */
        printk("Error while formatting block %d\n", block);
        return -1;
    } else {
        /* write erase info */
        nb_erases = le32_to_cpu(h1.WearInfo);
        nb_erases++;
        /* wrap (almost impossible with current flashs) or free block */
        if (nb_erases == 0)
            nb_erases = 1;
        h1.WearInfo = le32_to_cpu(nb_erases);
        if (MTD_WRITEOOB(s->mtd, block * s->EraseSize + SECTORSIZE + 8, 8,
                         &retlen, (char *)&h1) < 0)
            return -1;
        return 0;
    }
}

/* Check that each sector of a chain is correct. Mark as 'IGNORE' each
   incorrect sector. This check is only done if the chain was being
   folded when NFTL was interrupted. */
static void check_sectors_in_chain(struct NFTLrecord *s, unsigned int first_block)
{
    unsigned int block, i, status;
    struct nftl_bci h0;
    int sectors_per_block, retlen;

    sectors_per_block = s->EraseSize / SECTORSIZE;
    block = first_block;
    for(;;) {
        for(i=0;i<sectors_per_block;i++) {
            if (MTD_READOOB(s->mtd, block * s->EraseSize + i * SECTORSIZE, 8, &retlen, 
                            (char *)&h0) < 0)
                status = SECTOR_IGNORE;
            else
                status = h0.Status | h0.Status1;
            switch(status) {
            case SECTOR_FREE:
                /* verify that the sector is really free. If not, mark
                   as ignore */
                if (memcmpb(&h0, 0xff, 8) != 0 ||
                    check_free_sectors(s, block * s->EraseSize + i * SECTORSIZE, 
                                       SECTORSIZE, 0) != 0) {
                    printk("Incorrect free sector %d in block %d: marking it as ignored\n",
                           i, block);
                    /* sector not free : ignore it */
                    h0.Status = SECTOR_IGNORE;
                    h0.Status1 = SECTOR_IGNORE;
                    MTD_WRITEOOB(s->mtd, block * s->EraseSize + i * SECTORSIZE, 8, &retlen, 
                                 (char *)&h0);
                }
                break;
            default:
                break;
            }
        }
        block = s->ReplUnitTable[block];
        if (!(block == BLOCK_NIL || block < s->nb_blocks))
            printk("incorrect ReplUnitTable[] : %d\n", block);
        if (block == BLOCK_NIL ||
            block >= s->nb_blocks)
            break;
    }
}

/* Estimate chain length */
static int calc_chain_length(struct NFTLrecord *s, unsigned int first_block)
{
    unsigned int length, block;
    
    length = 0;
    block = first_block;
    for(;;) {
        length++;
        /* avoid infinite loops, although this is guaranted not to
           happen because of the previous checks */
        if (length >= s->nb_blocks) {
            printk("nftl: length too long %d !\n", length);
            break;
        }
        block = s->ReplUnitTable[block];
        if (!(block == BLOCK_NIL || block < s->nb_blocks))
            printk("incorrect ReplUnitTable[] : %d\n", block);
        if (block == BLOCK_NIL ||
            block >= s->nb_blocks)
            break;
    }
    return length;
}

/* Format an invalid chain. It is not stricly correct to begin from
   the first block of the chain because if we stop the code, we may
   see again a valid chain if there was a first_block flag in a block
   inside it. But is it really a problem ? */
static void format_chain(struct NFTLrecord *s, unsigned int first_block)
{
    unsigned int block, block1;

    printk("Formatting chain at block %d\n", first_block);

    block = first_block;
    for(;;) {
        block1 = s->ReplUnitTable[block];
        printk("Formatting block %d\n", block);
        if (NFTL_formatblock(s, block) < 0) {
            /* cannot format !!!! */
            s->ReplUnitTable[block] = BLOCK_RESERVED;
        } else {
            s->ReplUnitTable[block] = BLOCK_FREE;
        }
        block = block1;
        if (!(block == BLOCK_NIL || block < s->nb_blocks))
            printk("incorrect ReplUnitTable[] : %d\n", block);
        if (block == BLOCK_NIL ||
            block >= s->nb_blocks)
            break;
    }
}

/* Verify that a block is free in the NFTL sense (valid erase mark) or
   totally free (only 0xff).  */
static int check_and_mark_free_block(struct NFTLrecord *s, int block)
{
    struct nftl_uci1 h1;
    unsigned int erase_mark;
    int i, retlen;
    unsigned char buf[SECTORSIZE];

    /* check erase mark. */
    if (MTD_READOOB(s->mtd, block * s->EraseSize + SECTORSIZE + 8, 8, 
                    &retlen, (char *)&h1) < 0)
        return -1;
    erase_mark = le16_to_cpu ((h1.EraseMark | h1.EraseMark1));
    if (erase_mark != ERASE_MARK) {
        /* if no erase mark, the block must be totally free. This is
           possible in two cases : empty filsystem or interrupted erase (very unlikely) */
        if (check_free_sectors (s, block * s->EraseSize, s->EraseSize, 1) != 0)
            return -1;
        /* free block : write erase mark */
        h1.EraseMark = cpu_to_le16(ERASE_MARK);
        h1.EraseMark1 = cpu_to_le16(ERASE_MARK);
        h1.WearInfo = cpu_to_le32(0);
        if (MTD_WRITEOOB(s->mtd, block * s->EraseSize + SECTORSIZE + 8, 8, 
                         &retlen, (char *)&h1) < 0)
            return -1;
    } else {
        /* if erase mark present, need to skip it when doing check */
        for(i=0;i<s->EraseSize;i+=SECTORSIZE) {
            /* check free sector */
            if (check_free_sectors (s, block * s->EraseSize + i, SECTORSIZE, 0) != 0)
                return -1;
            if (MTD_READOOB(s->mtd, block * s->EraseSize + i, 16, &retlen, buf) < 0)
                return -1;
            if (i == SECTORSIZE) {
                /* skip erase mark */
                if (memcmpb(buf, 0xff, 8))
                    return -1;
            } else {
                if (memcmpb(buf, 0xff, 16))
                    return -1;
            }
        }
    }
    return 0;
}

/* WARNING: return 0 if read error */
static int get_fold_mark(struct NFTLrecord *s, unsigned int block)
{
    struct nftl_uci2 h2;
    int retlen;

    if (MTD_READOOB(s->mtd, block * s->EraseSize + 2 * SECTORSIZE + 8, 8,
                    &retlen, (char *)&h2) < 0)
        return 0;
    
    return le16_to_cpu((h2.FoldMark | h2.FoldMark1));
}

int NFTL_mount(struct NFTLrecord *s)
{
    int i;
    unsigned int first_logical_block, logical_block, rep_block, nb_erases, erase_mark;
    unsigned int block, first_block, is_first_block;
    int chain_length, do_format_chain;
    struct nftl_uci0 h0;
    struct nftl_uci1 h1;
    int retlen;

    /* XXX: will be suppressed */
    s->lastEUN = s->nb_blocks -1;
    /* memory alloc */
    s->EUNtable = kmalloc(s->nb_blocks * sizeof(u16), GFP_KERNEL);
    s->ReplUnitTable = kmalloc(s->nb_blocks * sizeof(u16), GFP_KERNEL);
    if (!s->EUNtable ||
        !s->ReplUnitTable) {
    fail:
        if (s->EUNtable)
            kfree(s->EUNtable);
        if (s->ReplUnitTable)
            kfree(s->ReplUnitTable);
        return -1;
    }

    /* mark all blocks as potentially containing data */
    for(i=0;i<s->nb_blocks;i++) { 
        s->ReplUnitTable[i] = BLOCK_NOTEXPLORED;
    }

    if (find_boot_record(s) < 0) {
        printk("Could not find valid boot record\n");
        goto fail;
    }
    
    /* mark the bios blocks as reserved */
    for(i=0;i<s->nb_boot_blocks;i++)
        s->ReplUnitTable[i] = BLOCK_RESERVED;
    
    /* also mark the boot records blocks as reserved */
    if (s->MediaUnit != BLOCK_NIL)
        s->ReplUnitTable[s->MediaUnit] = BLOCK_RESERVED;
    if (s->SpareMediaUnit != BLOCK_NIL)
        s->ReplUnitTable[s->SpareMediaUnit] = BLOCK_RESERVED;
    
    /* init the logical to physical table */
    for(i=0;i<s->nb_blocks;i++) {
        s->EUNtable[i] = BLOCK_NIL;
    }
    
    /* first pass : explore each block chain */

    first_logical_block = 0;
    for(first_block=0;first_block<s->nb_blocks;first_block++) {
        /* if the block was not already explored, we can look at it */
        if (s->ReplUnitTable[first_block] == BLOCK_NOTEXPLORED) {
            block = first_block;
            chain_length = 0;
            do_format_chain = 0;
            for(;;) {
                /* read the block header. If error, we format the chain */
                if (MTD_READOOB(s->mtd, block * s->EraseSize + 8, 8, 
                                &retlen, (char *)&h0) < 0 ||
                    MTD_READOOB(s->mtd, block * s->EraseSize + SECTORSIZE + 8, 8, 
                                &retlen, (char *)&h1) < 0) {
                    s->ReplUnitTable[block] = BLOCK_NIL;
                    do_format_chain = 1;
                    break;
                }
                
                logical_block = le16_to_cpu ((h0.VirtUnitNum | h0.SpareVirtUnitNum));
                rep_block = le16_to_cpu ((h0.ReplUnitNum | h0.SpareReplUnitNum));
                nb_erases = le32_to_cpu (h1.WearInfo);
                erase_mark = le16_to_cpu ((h1.EraseMark | h1.EraseMark1));

                is_first_block = !(logical_block >> 15);
                logical_block = logical_block & 0x7fff;
                
                /* invalid/free block test */
                if (erase_mark != ERASE_MARK ||
                    logical_block >= s->nb_blocks) {
                    
                    if (chain_length == 0) {
                        /* if not currently in a chain, we can handle it safely */
                        if (check_and_mark_free_block(s, block) < 0) {
                            /* not really free: format it */
                            printk("Formatting block %d\n", block);
                            if (NFTL_formatblock(s, block) < 0) {
                                /* could not format: reserve the block */
                                s->ReplUnitTable[block] = BLOCK_RESERVED;
                            } else {
                                s->ReplUnitTable[block] = BLOCK_FREE;
                            }
                        } else {
                            /* free block: mark it */
                            s->ReplUnitTable[block] = BLOCK_FREE;
                        }
                        /* directly examine the next block. */
                        goto examine_ReplUnitTable;
                    } else {
                        /* the bloc was in a chain : this is bad. We
                           must format all the chain */
                        printk("Block %d: free but referenced in chain %d\n",
                               block, first_block);
                        s->ReplUnitTable[block] = BLOCK_NIL;
                        do_format_chain = 1;
                        break;
                    }
                }

                /* we accept only first blocks here */
                if (chain_length == 0) {
                    /* this block is not the first block in chain :
                       ignore it, it will be included in a chain
                       later, or marked as not explored */
                    if (!is_first_block)
                        goto examine_ReplUnitTable;
                    first_logical_block = logical_block;
                } else {
                    if (logical_block != first_logical_block) {
                        printk("Block %d: incorrect logical block: %d expected: %d\n", 
                               block, logical_block, first_logical_block);
                        /* the chain is incorrect : we must format it,
                           but we need to read it completly */
                        do_format_chain = 1;
                    }
                    if (is_first_block) {
                        /* we accept that a block is marked as first
                           block while being last block in a chain
                           only if the chain is being folded */
                        if (get_fold_mark(s, block) != FOLD_MARK_IN_PROGRESS ||
                            rep_block != 0xffff) {
                            printk("Block %d: incorrectly marked as first block in chain\n",
                                   block);
                            /* the chain is incorrect : we must format it,
                               but we need to read it completly */
                            do_format_chain = 1;
                        } else {
                            printk("Block %d: folding in progress - ignoring first block flag\n",
                                   block);
                        }
                    }
                }
                chain_length++;
                if (rep_block == 0xffff) {
                    /* no more blocks after */
                    s->ReplUnitTable[block] = BLOCK_NIL;
                    break;
                } else if (rep_block >= s->nb_blocks) {
                    printk("Block %d: referencing invalid block %d\n", 
                           block, rep_block);
                    do_format_chain = 1;
                    s->ReplUnitTable[block] = BLOCK_NIL;
                    break;
                } else if (s->ReplUnitTable[rep_block] != BLOCK_NOTEXPLORED) {
                    /* same problem as previous 'is_first_block' test:
                       we accept that the last block of a chain has
                       the first_block flag set if folding is in
                       progress. We handle here the case where the
                       last block appeared first */
                    if (s->ReplUnitTable[rep_block] == BLOCK_NIL &&
                        s->EUNtable[first_logical_block] == rep_block &&
                        get_fold_mark(s, first_block) == FOLD_MARK_IN_PROGRESS) {
                        /* EUNtable[] will be set after */
                        printk("Block %d: folding in progress - ignoring first block flag\n",
                               rep_block);
                        s->ReplUnitTable[block] = rep_block;
                        s->EUNtable[first_logical_block] = BLOCK_NIL;
                    } else {
                        printk("Block %d: referencing block %d already in another chain\n", 
                               block, rep_block);
                        /* XXX: should handle correctly fold in progress chains */
                        do_format_chain = 1;
                        s->ReplUnitTable[block] = BLOCK_NIL;
                    }
                    break;
                } else {
                    /* this is OK */
                    s->ReplUnitTable[block] = rep_block;
                    block = rep_block;
                }
            }

            /* the chain was completely explored. Now we can decide
               what to do with it */
            if (do_format_chain) {
                /* invalid chain : format it */
                format_chain(s, first_block);
            } else {
                unsigned int first_block1, chain_to_format, chain_length1;
                int fold_mark;

                /* valid chain : get foldmark */
                fold_mark = get_fold_mark(s, first_block);
                if (fold_mark == 0) {
                    /* cannot get foldmark : format the chain */
                    printk("Could read foldmark at block %d\n", first_block);
                    format_chain(s, first_block);
                } else {
                    if (fold_mark == FOLD_MARK_IN_PROGRESS)
                        check_sectors_in_chain(s, first_block);

                    /* now handle the case where we find two chains at the
                       same virtual address : we select the longer one,
                       because the shorter one is the one which was being
                       folded if the folding was not done in place */
                    first_block1 = s->EUNtable[first_logical_block];
                    if (first_block1 != BLOCK_NIL) {
                        /* XXX: what to do if same length ? */
                        chain_length1 = calc_chain_length(s, first_block1);
                        printk("Two chains at blocks %d (len=%d) and %d (len=%d)\n", 
                               first_block1, chain_length1, first_block, chain_length);

                        if (chain_length >= chain_length1) {
                            chain_to_format = first_block1;
                            s->EUNtable[first_logical_block] = first_block;
                        } else {
                            chain_to_format = first_block;
                        }
                        format_chain(s, chain_to_format);
                    } else {
                        s->EUNtable[first_logical_block] = first_block;
                    }
                }
            }
        }
    examine_ReplUnitTable:
    }

    /* second pass to format unreferenced blocks  and init free block count */
    s->numfreeEUNs = 0;
    s->LastFreeEUN = BLOCK_NIL;
    
    for(block=0;block<s->nb_blocks;block++) {
        if (s->ReplUnitTable[block] == BLOCK_NOTEXPLORED) {
            printk("Unreferenced block %d, formatting it\n", block);
            if (NFTL_formatblock(s, block) < 0)
                s->ReplUnitTable[block] = BLOCK_RESERVED;
            else
                s->ReplUnitTable[block] = BLOCK_FREE;
        }
        if (s->ReplUnitTable[block] == BLOCK_FREE) {
            s->numfreeEUNs++;
            s->LastFreeEUN = block;
        }
    }
    return 0;
}
