/*
 * Philips UDA1341 Audio Device Driver for SA1100 Linux
 *
 * Copyright (c) 2000 Nicolas Pitre <nico@cam.org>
 *
 * Portions are Copyright (C) 2000 Lernout & Hauspie Speech Products, N.V.

 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License.
 *
 * History:
 *
 * 2000-05-21	Nicolas Pitre	Initial release.
 *
 * 2000-06-10	Erik Bunce	Add initial poll support.
 * 
 * 2000-07-??	George France	Bitsy support.
 *
 * 2000-08-19	Erik Bunce	More inline w/ OSS API and UDA1341 docs
 * 				including fixed AGC and audio source handling
 *
 * 2000-08-22	Nicolas Pitre	Removed all DMA stuff. Now using the 
 * 				generic SA1100 DMA interface.
 *
 * 2000-09-04	John Dorsey     SA-1111 Serial Audio Controller support.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/interrupt.h>
#include <linux/errno.h>
#include <linux/sound.h>
#include <linux/soundcard.h>

#include <linux/pm.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/semaphore.h>
#include <asm/dma.h>


#undef DEBUG
#ifdef DEBUG
#define DPRINTK( x... )  printk( ##x )
#else
#define DPRINTK( x... )
#endif


/* 
 * Definitions 
 */

#define AUDIO_NAME		"UDA1341"
#define AUDIO_NAME_VERBOSE	"UDA1341 audio driver"
#define AUDIO_VERSION_STRING	"version 0.6"

#define AUDIO_FMT_MASK		(AFMT_S16_LE)
#define AUDIO_FMT_DEFAULT	(AFMT_S16_LE)
#define AUDIO_CHANNELS_DEFAULT	2
#define AUDIO_RATE_DEFAULT	44100
#define AUDIO_NBFRAGS_DEFAULT	8
#define AUDIO_FRAGSIZE_DEFAULT	8192


/*
 * GPIO based L3 bus support.
 *
 * This provides control of Philips L3 type devices. 
 * GPIO lines are used for clock, data and mode pins.
 *
 * Note: On Assabet the L3 pins are shared with I2C devices. This should not 
 * present any problems as long as an I2C start sequence is not generated. 
 * This is defined as a 1->0 transition on the data lines when the clock is 
 * high. It is critical this code only allow data transitions when the clock
 * is low. This is always legal in L3.
 *
 * The IIC interface requires the clock and data pin to be LOW when idle. We
 * must make sure we leave them in this state.
 *
 * It appears the read data is generated on the falling edge of the clock
 * and should be held stable during the clock high time.
 */

/* 
 * L3 bus GPIO pin definitions
 */

/* FIXME: should be made of variables instead... */
#ifdef CONFIG_SA1100_ASSABET
#define L3_DataPin	GPIO_GPIO(15)
#define L3_ClockPin	GPIO_GPIO(18)
#define L3_ModePin	GPIO_GPIO(17)
#endif
#ifdef CONFIG_SA1100_BITSY
#define L3_DataPin	GPIO_BITSY_L3_DATA
#define L3_ClockPin	GPIO_BITSY_L3_CLOCK
#define L3_ModePin	GPIO_BITSY_L3_MODE
#endif


/* 
 * L3 setup and hold times (expressed in us)
 */
#define L3_DataSetupTime 1	/* 190 ns */
#define L3_DataHoldTime  1	/*  30 ns */
#define L3_ModeSetupTime 1	/* 190 ns */
#define L3_ModeHoldTime  1	/* 190 ns */
#define L3_ClockHighTime 1	/* 250 ns (min is 64*fs, 35us @ 44.1 Khz) */
#define L3_ClockLowTime  1	/* 250 ns (min is 64*fs, 35us @ 44.1 Khz) */
#define L3_HaltTime      1	/* 190 ns */

/*
 * Grab control of the IIC/L3 shared pins
 */
static inline void L3_acquirepins(void)
{
	if (machine_is_assabet()) {
		GPSR = (L3_ClockPin | L3_DataPin);
		GPDR |= (L3_ClockPin | L3_DataPin);
	}
}

/*
 * Release control of the IIC/L3 shared pins
 */
static inline void L3_releasepins(void)
{
	if (machine_is_assabet()) {
		GPDR &= ~(L3_ClockPin | L3_DataPin);
		GPCR = (L3_ClockPin | L3_DataPin);
	}
}

/*
 * Initialize the interface
 */
static void __init L3_init(void)
{
  	if(!machine_has_neponset()){
		GAFR &= ~(L3_DataPin | L3_ClockPin | L3_ModePin);
		GPSR = (L3_ModePin | L3_ClockPin | L3_DataPin);
		GPDR |= (L3_ModePin | L3_ClockPin | L3_DataPin);
		L3_releasepins();
	}
}

/*
 * Get a bit. The clock is high on entry and on exit. Data is read after
 * the clock low time has expired.
 */
static inline int L3_getbit(void)
{
	int data;

	GPCR = L3_ClockPin;
	udelay(L3_ClockLowTime);

	data = (GPLR & L3_DataPin) ? 1 : 0;

	GPSR = L3_ClockPin;
	udelay(L3_ClockHighTime);

	return data;
}

/*
 * Send a bit. The clock is high on entry and on exit. Data is sent only
 * when the clock is low (I2C compatibility).
 */
static inline void L3_sendbit(int bit)
{
	GPCR = L3_ClockPin;

	if (bit & 1)
		GPSR = L3_DataPin;
	else
		GPCR = L3_DataPin;

	/* Assumes L3_DataSetupTime < L3_ClockLowTime */
	udelay(L3_ClockLowTime);

	GPSR = L3_ClockPin;
	udelay(L3_ClockHighTime);
}

/*
 * Send a byte. The mode line is set or pulsed based on the mode sequence
 * count. The mode line is high on entry and exit. The mod line is pulsed
 * before the second data byte and before ech byte thereafter.
 */
static void L3_sendbyte(char data, int mode)
{
	int i;

	L3_acquirepins();

	switch (mode) {
	case 0:		/* Address mode */
		GPCR = L3_ModePin;
		break;
	case 1:		/* First data byte */
		break;
	default:		/* Subsequent bytes */
		GPCR = L3_ModePin;
		udelay(L3_HaltTime);
		GPSR = L3_ModePin;
		break;
	}

	udelay(L3_ModeSetupTime);

	for (i = 0; i < 8; i++)
		L3_sendbit(data >> i);

	if (mode == 0)		/* Address mode */
		GPSR = L3_ModePin;

	udelay(L3_ModeHoldTime);

	L3_releasepins();
}

/*
 * Get a byte. The mode line is set or pulsed based on the mode sequence
 * count. The mode line is high on entry and exit. The mod line is pulsed
 * before the second data byte and before each byte thereafter. This
 * function is never valid with mode == 0 (address cycle) as the address
 * is always sent on the bus, not read.
 */
static char L3_getbyte(int mode)
{
	char data = 0;
	int i;

	L3_acquirepins();
	GPDR &= ~(L3_DataPin);

	switch (mode) {
	case 0:		/* Address mode - never valid */
		break;
	case 1:		/* First data byte */
		break;
	default:		/* Subsequent bytes */
		GPCR = L3_ModePin;
		udelay(L3_HaltTime);
		GPSR = L3_ModePin;
		break;
	}

	udelay(L3_ModeSetupTime);

	for (i = 0; i < 8; i++)
		data |= (L3_getbit() << i);

	udelay(L3_ModeHoldTime);

	L3_releasepins();

	return data;
}

/*
 * Write data to a device on the L3 bus. The address is passed as well as
 * the data and length. The length written is returned. The register space
 * is encoded in the address (low two bits are set and device address is
 * in the upper 6 bits).
 */
static int L3_write(char addr, char *data, int len)
{
	int bytes = len;

	DPRINTK("%s(0x%x, %d)\n", __FUNCTION__, addr, len);

	if(machine_has_neponset()){  /* SA-1111 L3 Control Bus */
#ifdef CONFIG_SA1111
		if( len > 1 ){
		  	SACR1 |= SACR1_L3MB;
			while( (len--) > 1 ){
			  	L3_CAR = addr;
			  	L3_CDR = *data++;
			  	while((SASR0 & SASR0_L3WD) == 0)
				  	mdelay(1);
				SASCR = SASCR_DTS;
			}
		}
		SACR1 &= ~SACR1_L3MB;
	  	L3_CAR = addr;
		L3_CDR = *data;
		while((SASR0 & SASR0_L3WD) == 0)
		  	mdelay(1);
		SASCR = SASCR_DTS;
#endif
	} else {

	  	int mode = 0;

		L3_sendbyte(addr, mode++);
		while(len--)
		  	L3_sendbyte(*data++, mode++);
	}

	return bytes;
}

/*
 * Read data from a device on the L3 bus. The address is passed as well as
 * the data and length. The length read is returned. The register space
 * is encoded in the address (low two bits are set and device address is
 * in the upper 6 bits).
 */
static int L3_read(char addr, char *data, int len)
{
	int bytes = len;

	DPRINTK("%s(0x%x, %d)\n", __FUNCTION__, addr, len);

	if(machine_has_neponset()){  /* SA-1111 L3 Control Bus */
#ifdef CONFIG_SA1111
		if( len > 1 ){
		  	SACR1 |= SACR1_L3MB;
			while( (len--) > 1 ){
			  	L3_CAR = addr;
			  	while((SASR0 & SASR0_L3RD) == 0)
				  	mdelay(1);
				*data++ = L3_CDR;
				SASCR = SASCR_RDD;
			}
		}
		SACR1 &= ~SACR1_L3MB;
		L3_CAR = addr;
		while((SASR0 & SASR0_L3RD) == 0)
		  	mdelay(1);
		*data = L3_CDR;
		SASCR = SASCR_RDD;
#endif
	} else {

		int mode = 0;
		
		L3_sendbyte(addr, mode++);
		while(len--)
		  	*data++ = L3_getbyte(mode++);
	}

	return bytes;
}


/*
 * UDA1341 L3 address and command types
 */
#define UDA1341_L3Addr	5
#define UDA1341_DATA0	0
#define UDA1341_DATA1	1
#define UDA1341_STATUS	2


/* 
 * UDA1341 internal state variables.  
 * Those are initialized according to what it should be after a reset.
 */

/* UDA1341 status settings */

#define UDA_STATUS0_IF_I2S      0
#define UDA_STATUS0_IF_LSB16    1
#define UDA_STATUS0_IF_LSB18    2
#define UDA_STATUS0_IF_LSB20    3
#define UDA_STATUS0_IF_MSB      4
#define UDA_STATUS0_IF_MSB16    5
#define UDA_STATUS0_IF_MSB18    6
#define UDA_STATUS0_IF_MSB20    7

#define UDA_STATUS0_SC_512FS   0
#define UDA_STATUS0_SC_384FS   1
#define UDA_STATUS0_SC_256FS   2

static struct {
	u_int DC_filter:1;	/* DC filter */
	u_int input_fmt:3;	/* data input format */
	u_int system_clk:2;	/* system clock frequency */
	u_int reset:1;		/* reset */
	const u_int select:1;	/* must be set to 0 */
} STATUS_0 = {
0, UDA_STATUS0_IF_LSB16, UDA_STATUS0_SC_256FS, 0, 0};

static struct {
	u_int DAC_on:1;		/* DAC powered */
	u_int ADC_on:1;		/* ADC powered */
	u_int double_speed:1;	/* double speed playback */
	u_int DAC_pol:1;	/* polarity of DAC */
	u_int ADC_pol:1;	/* polarity of ADC */
	u_int ADC_gain:1;	/* gain of ADC */
	u_int DAC_gain:1;	/* gain of DAC */
	const u_int select:1;	/* must be set to 1 */
} STATUS_1 = {
1, 1, 0, 0, 0, 0, 0, 1};

/* UDA1341 direct control settings */

static struct {
	u_int volume:6;		/* volume control */
	const u_int select:2;	/* must be set to 0 */
} DATA0_0 = {
0, 0};

static struct {
	u_int treble:2;
	u_int bass:4;
	const u_int select:2;	/* must be set to 1 */
} DATA0_1 = {
0, 0, 1};

static struct {
	u_int mode:2;		/* mode switch */
	u_int mute:1;
	u_int deemphasis:2;
	u_int peak_detect:1;
	const u_int select:2;	/* must be set to 2 */
} DATA0_2 = {
0, 0, 0, 1, 2};

/* DATA0 extended programming registers */

static struct {
	const u_int ext_addr:3;	/* must be set to 0 */
	const u_int select1:5;	/* must be set to 24 */
	u_int ch1_gain:5;	/* mixer gain channel 1 */
	const u_int select2:3;	/* must be set to 7 */
} DATA0_ext0 = {
0, 24, 4, 7};

static struct {
	const u_int ext_addr:3;	/* must be set to 1 */
	const u_int select1:5;	/* must be set to 24 */
	u_int ch2_gain:5;	/* mixer gain channel 2 */
	const u_int select2:3;	/* must be set to 7 */
} DATA0_ext1 = {
1, 24, 4, 7};

static struct {
	const u_int ext_addr:3;	/* must be set to 2 */
	const u_int select1:5;	/* must be set to 24 */
	u_int mixer_mode:2;
	u_int mic_level:3;	/* MIC sensitivity level */
	const u_int select2:3;	/* must be set to 7 */
} DATA0_ext2 = {
2, 24, 0, 1, 7};

static struct {
	const u_int ext_addr:3;	/* must be set to 4 */
	const u_int select1:5;	/* must be set to 24 */
	u_int ch2_igain_l:2;	/* input amplifier gain channel 2 (bits 1-0) */
	const u_int reserved:2;	/* must be set to 0 */
	u_int AGC_ctrl:1;	/* AGC control */
	const u_int select2:3;	/* must be set to 7 */
} DATA0_ext4 = {
4, 24, 0 & 3, 0, 0, 7};

static struct {
	const u_int ext_addr:3;	/* must be set to 5 */
	const u_int select1:5;	/* must be set to 24 */
	u_int ch2_igain_h:5;	/* input amplifier gain channel 2 (bits 6-2) */
	const u_int select2:3;	/* must be set to 7 */
} DATA0_ext5 = {
5, 24, 0 >> 2, 7};

static struct {
	const u_int ext_addr:3;	/* must be set to 6 */
	const u_int select1:5;	/* must be set to 24 */
	u_int AGC_level:2;	/* AGC output level */
	u_int AGC_const:3;	/* AGC time constant */
	const u_int select2:3;	/* must be set to 7 */
} DATA0_ext6 = {
6, 24, 0, 0, 7};

static struct {
	u_int peak:6;		/* peak level value */
} DATA1 = { 0 };


/* 
 * Buffer Management
 */

typedef struct {
	int size;		/* buffer size */
	char *start;		/* points to actual buffer */
	dma_addr_t dma_addr;	/* physical buffer address */
	struct semaphore sem;	/* down before touching the buffer */
	int master;		/* master owner for buffer allocation */
} audio_buf_t;

typedef struct {
	audio_buf_t *buffers;	/* pointer to audio buffer structures */
	audio_buf_t *buf;	/* current buffer used by read/write */
	u_int buf_idx;		/* index for the pointer above... */
	u_int fragsize;		/* fragment i.e. buffer size */
	u_int nbfrags;		/* nbr of fragments i.e. buffers */
	dmach_t dma_ch;		/* DMA channel ID */
} audio_stream_t;

static audio_stream_t output_stream;
static audio_stream_t input_stream;

#define NEXT_BUF(_s_,_b_) { \
	(_s_)->_b_##_idx++; \
	(_s_)->_b_##_idx %= (_s_)->nbfrags; \
	(_s_)->_b_ = (_s_)->buffers + (_s_)->_b_##_idx; }


/* Current specs for incoming audio data */
static u_int audio_rate;
static int audio_channels;
static int audio_fmt;
static u_int audio_fragsize;
static u_int audio_nbfrags;


static int audio_refcount;	/* nbr of concurrent open() */
static int audio_dev_dsp;	/* registered ID for DSP device */
static int audio_dev_mixer;	/* registered ID for mixer device */

#ifdef CONFIG_PM
static struct pm_dev *audio_pm_dev; 
#endif

static void audio_uda1341_reset(void);


/* 
 * This function frees all buffers 
 */

static void audio_clear_buf(audio_stream_t * s)
{
	DPRINTK("audio_clear_buf\n");

	/* ensure DMA won't run anymore */
	sa1100_dma_flush_all(s->dma_ch);

	if (s->buffers) {
		int frag;
		for (frag = 0; frag < s->nbfrags; frag++) {
			if (!s->buffers[frag].master)
				continue;
			consistent_free(s->buffers[frag].start);
		}
		kfree(s->buffers);
		s->buffers = NULL;
	}

	s->buf_idx = 0;
	s->buf = NULL;
}


/* This function allocates the buffer structure array and buffer data space
 * according to the current number of fragments and fragment size.
 */

static int audio_setup_buf(audio_stream_t * s)
{
	int frag;
	int dmasize = 0;
	char *dmabuf = 0;
	dma_addr_t dmaphys = 0;

	if (s->buffers)
		return -EBUSY;

	s->nbfrags = audio_nbfrags;
	s->fragsize = audio_fragsize;

	s->buffers = (audio_buf_t *)
	    kmalloc(sizeof(audio_buf_t) * s->nbfrags, GFP_KERNEL);
	if (!s->buffers)
		goto err;
	memset(s->buffers, 0, sizeof(audio_buf_t) * s->nbfrags);

	for (frag = 0; frag < s->nbfrags; frag++) {
		audio_buf_t *b = &s->buffers[frag];

		/*
		 * Let's allocate non-cached memory for DMA buffers.
		 * We try to allocate all memory at once.
		 * If this fails (a common reason is memory fragmentation), 
		 * then we allocate more smaller buffers.
		 */
		if (!dmasize) {
			dmasize = (s->nbfrags - frag) * s->fragsize;
			do {
			  	dmabuf = consistent_alloc(GFP_KERNEL|GFP_DMA,
							  dmasize,
							  &dmaphys);
				if (!dmabuf)
					dmasize -= s->fragsize;
			} while (!dmabuf && dmasize);
			if (!dmabuf)
				goto err;
			b->master = 1;
		}

		b->start = dmabuf;
		b->dma_addr = dmaphys;
		sema_init(&b->sem, 1);
		DPRINTK("buf %d: start %p dma %p\n", frag, b->start,
			b->dma_addr);

		dmabuf += s->fragsize;
		dmaphys += s->fragsize;
		dmasize -= s->fragsize;
	}

	s->buf_idx = 0;
	s->buf = &s->buffers[0];

	return 0;

      err:
	printk(AUDIO_NAME ": unable to allocate audio memory\n ");
	audio_clear_buf(s);
	return -ENOMEM;
}


/*
 * DMA callback functions
 */

static void audio_dmaout_done_callback(void *buf_id, int size)
{
	audio_buf_t *b = (audio_buf_t *) buf_id;
	/* 
	 * Current buffer is sent: wake up any process waiting for it.
	 */
	up(&b->sem);

#define CHROME 1 /* What else are you going to do with 32 LEDs? */
#ifdef CHROME
	if (machine_is_assabet() && machine_has_neponset()) {
#ifdef CONFIG_ASSABET_NEPONSET
	  	L3_read( (UDA1341_L3Addr<<2)|UDA1341_DATA1, 
			 (char*)&DATA1, 1 );
		LEDS = (1 << (DATA1.peak >> 1)) - 1;
#endif
	}
#endif
}

static void audio_dmain_done_callback(void *buf_id, int size)
{
	audio_buf_t *b = (audio_buf_t *) buf_id;
	/* 
	 * Current buffer is full: set its size and wake up any 
	 * process waiting for it.
	 */
	b->size = size;
	up(&b->sem);
}


static int audio_sync(struct file *file)
{
	audio_stream_t *s = &output_stream;
	audio_buf_t *b = s->buf;

	DPRINTK("audio_sync\n");

	if (!s->buffers)
		return 0;

	/* Send half-full buffers */
	if (b->size != 0) {
		down(&b->sem);
		sa1100_dma_queue_buffer(s->dma_ch, (void *) b,
					b->dma_addr, b->size);
		b->size = 0;
		NEXT_BUF(s, buf);
	}

	/*
	 * Let's wait for the last buffer we sent i.e. the one before the
	 * current buf_idx.  When we acquire the semaphore, this means either:
	 * - DMA on the buffer completed or 
	 * - the buffer was already free thus nothing else to sync.
	 */
	b = s->buffers + ((s->nbfrags + s->buf_idx - 1) % s->nbfrags);
	if (down_interruptible(&b->sem))
		return -EINTR;
	up(&b->sem);
	return 0;
}


static int audio_write(struct file *file, const char *buffer,
		       size_t count, loff_t * ppos)
{
	const char *buffer0 = buffer;
	audio_stream_t *s = &output_stream;
	int chunksize, ret = 0;

	DPRINTK("audio_write: count=%d\n", count);

	if (!s->buffers && audio_setup_buf(s))
		return -ENOMEM;

	/* be sure to have a full sample byte count */
	count &= ~0x03;

	while (count > 0) {
		audio_buf_t *b = s->buf;

		/* Wait for a buffer to become free */
		if (file->f_flags & O_NONBLOCK) {
			ret = -EAGAIN;
			if (down_trylock(&b->sem))
				break;
		} else {
			ret = -ERESTARTSYS;
			if (down_interruptible(&b->sem))
				break;
		}

		/* Feed the current buffer */
		chunksize = s->fragsize - b->size;
		if (chunksize > count)
			chunksize = count;
		DPRINTK("write %d to %d\n", chunksize, s->buf_idx);
		if (copy_from_user(b->start + b->size, buffer, chunksize)) {
			up(&b->sem);
			return -EFAULT;
		}
		buffer += chunksize;
		count -= chunksize;
		b->size += chunksize;
		if (b->size < s->fragsize) {
			up(&b->sem);
			break;
		}

		/* Send current buffer to dma */
		sa1100_dma_queue_buffer(s->dma_ch, (void *) b,
					b->dma_addr, b->size);
		b->size = 0;	/* indicate that the buffer has been sent */
		NEXT_BUF(s, buf);
	}

	if ((buffer - buffer0))
		ret = buffer - buffer0;
	DPRINTK("audio_write: return=%d\n", ret);
	return ret;
}


static int audio_recording(audio_stream_t * s)
{
	int i;

	if (!s->buffers) {
		if (audio_setup_buf(s))
			return -ENOMEM;

		/*
		 * We must ensure there is an output stream at any time while 
		 * recording since this is how the UDA1341 gets its clock.
		 * So if there is no playback data to send, the output DMA will
		 * spin with all zeroes.
		 */
		if (!machine_is_assabet() || !machine_has_neponset())
		  	sa1100_dma_set_spin(output_stream.dma_ch,
					    (dma_addr_t) FLUSH_BASE_PHYS, 2048);

		/* 
		 * Since we just allocated all buffers, we must send them to
		 * the DMA code before receiving data.
		 */
		for (i = 0; i < s->nbfrags; i++) {
			audio_buf_t *b = s->buf;
			down(&b->sem);
			sa1100_dma_queue_buffer(s->dma_ch, (void *) b,
						b->dma_addr, s->fragsize);
			NEXT_BUF(s, buf);
		}
	}
	return 0;
}


static int audio_read(struct file *file, char *buffer,
		      size_t count, loff_t * ppos)
{
	char *buffer0 = buffer;
	audio_stream_t *s = &input_stream;
	int chunksize, ret = 0;

	DPRINTK("audio_read: count=%d\n", count);

	ret = audio_recording(s);
	if (ret)
		return ret;

	/* be sure to have a full sample byte count */
	count &= ~0x03;

	while (count > 0) {
		audio_buf_t *b = s->buf;

		/* Wait for a buffer to become full */
		if (file->f_flags & O_NONBLOCK) {
			ret = -EAGAIN;
			if (down_trylock(&b->sem))
				break;
		} else {
			ret = -ERESTARTSYS;
			if (down_interruptible(&b->sem))
				break;
		}

		/* Grab data from the current buffer */
		chunksize = b->size;
		if (chunksize > count)
			chunksize = count;
		DPRINTK("read %d from %d\n", chunksize, s->buf_idx);
		if (copy_to_user(buffer,
				 b->start + s->fragsize - b->size,
				 chunksize)) {
			up(&b->sem);
			return -EFAULT;
		}
		buffer += chunksize;
		count -= chunksize;
		b->size -= chunksize;
		if (b->size > 0) {
			up(&b->sem);
			break;
		}

		/* Make current buffer available for DMA again */
		sa1100_dma_queue_buffer(s->dma_ch, (void *) b,
					b->dma_addr, s->fragsize);
		NEXT_BUF(s, buf);
	}

	if ((buffer - buffer0))
		ret = buffer - buffer0;
	DPRINTK("audio_read: return=%d\n", ret);
	return ret;
}


static unsigned int audio_poll(struct file *file,
			       struct poll_table_struct *wait)
{
	unsigned int mask = 0;
	int i;
	int ret;

	DPRINTK("audio_poll(): mode=%s%s\n",
		(file->f_mode & FMODE_READ) ? "r" : "",
		(file->f_mode & FMODE_WRITE) ? "w" : "");

	if (file->f_mode & FMODE_READ) {
		/* Start audio input if not already active */
		ret = audio_recording(&input_stream);
		if (ret < 0)
			return ret;
	}

	if (file->f_mode & FMODE_WRITE) {
		if (!output_stream.buffers
		    && audio_setup_buf(&output_stream)) 
			return -ENOMEM;
	}

	poll_wait(file, &input_stream.buf->sem.wait, wait);
	poll_wait(file, &output_stream.buf->sem.wait, wait);

	if (file->f_mode & FMODE_READ) {
		for (i = 0; i < input_stream.nbfrags; i++) {
			if (atomic_read(&input_stream.buffers[i].sem.count) > 0)
				mask |= POLLIN | POLLRDNORM;
		}
	}
	if (file->f_mode & FMODE_WRITE) {
		for (i = 0; i < output_stream.nbfrags; i++) {
			if (atomic_read(&output_stream.buffers[i].sem.count) > 0)
				mask |= POLLOUT | POLLWRNORM;
		}
	}

	DPRINTK("audio_poll() returned mask of %s%s\n",
		(mask & POLLIN) ? "r" : "", 
		(mask & POLLOUT) ? "w" : "");

	return mask;
}


static loff_t audio_llseek(struct file *file, loff_t offset, int origin)
{
	return -ESPIPE;
}


static int mixer_ioctl(struct inode *inode, struct file *file,
		       uint cmd, ulong arg)
{
	long val = 0;

	/* dispatch based on command */
	switch (cmd) {
	case SOUND_MIXER_READ_DEVMASK:
		val = (SOUND_MASK_VOLUME |
		       SOUND_MASK_TREBLE |
		       SOUND_MASK_BASS |
		       SOUND_MASK_LINE1 |
		       SOUND_MASK_LINE2 | 
		       SOUND_MASK_MIC);
		put_user(val, (long *) arg);
		break;
	case SOUND_MIXER_READ_RECMASK:
		val = (SOUND_MASK_MIC |
		       SOUND_MASK_LINE1 |
		       SOUND_MASK_LINE2 | 
		       SOUND_MASK_LINE3);
		put_user(val, (long *) arg);
		break;
	case SOUND_MIXER_READ_STEREODEVS:
		put_user(0, (long *) arg);
		break;
	case SOUND_MIXER_READ_CAPS:
		val = SOUND_CAP_EXCL_INPUT;
		put_user(val, (long *) arg);
		break;
	case SOUND_MIXER_AGC:
		/* (as found in sb_mixer.c)
		 * Use ioctl(fd, SOUND_MIXER_AGC, &mode) to turn AGC off (0) or on (1).
		 */
		DATA0_ext4.AGC_ctrl = val ? 1 : 0;
		L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0,
			 (char *) &DATA0_ext4, 2);
		put_user(DATA0_ext4.AGC_ctrl, (long *) arg);
		break;
	case SOUND_MIXER_WRITE_RECSRC:
		get_user(val, (long *) arg);
		/* Recording source is selected by mixer_mode */
		switch (val) {
		case SOUND_MASK_LINE1:
			/* input channel 1 select */
			DATA0_ext2.mixer_mode = 1;
			break;
		case SOUND_MASK_LINE2:
			/* Double differential mode */
			DATA0_ext2.mixer_mode = 0;
			break;
		case SOUND_MASK_LINE3:
			/*
			 * digital mixer mode
			 * (input 1 x MA + input2 x MB)
			 */
			DATA0_ext2.mixer_mode = 3;
			break;
		case SOUND_MASK_MIC:
		default:
			/* Input channel 2 select */
			DATA0_ext2.mixer_mode = 2;
			break;
		}
		L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0,
			 (char *) &DATA0_ext2, 2);
		break;
	case SOUND_MIXER_READ_RECSRC:
		/* Recording source is specified by mixer_mode */
		switch (DATA0_ext2.mixer_mode) {
		case 0:
			/* Double differential mode */
			val = SOUND_MASK_LINE2;
			break;
		case 1:
			/* input channel 1 select */
			val = SOUND_MASK_LINE1;
			break;
		case 3:
			/*
			 * digital mixer mode
			 * (input 1 x MA + input2 x MB)
			 */
			val = SOUND_MASK_LINE3;
			break;
		case 2:
		default:
			/* Input channel 2 select */
			val = SOUND_MASK_MIC;
			break;
		}
		put_user(val, (long *) arg);
		break;

	case SOUND_MIXER_WRITE_VOLUME:
		get_user(val, (long *) arg);
		DATA0_0.volume = 63 - (((val & 0xff) + 1) * 63) / 100;
		L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0,
			 (char *) &DATA0_0, 1);
		break;
	case SOUND_MIXER_READ_VOLUME:
		val = ((63 - DATA0_0.volume) * 100) / 63;
		val |= val << 8;
		put_user(val, (long *) arg);
		break;
	case SOUND_MIXER_WRITE_TREBLE:
		get_user(val, (long *) arg);
		DATA0_1.treble = (((val & 0xff) + 1) * 3) / 100;
		L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0,
			 (char *) &DATA0_1, 1);
		break;
	case SOUND_MIXER_READ_TREBLE:
		val = (DATA0_1.treble * 100) / 3;
		val |= val << 8;
		put_user(val, (long *) arg);
		break;
	case SOUND_MIXER_WRITE_BASS:
		get_user(val, (long *) arg);
		DATA0_1.bass = (((val & 0xff) + 1) * 15) / 100;
		L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0,
			 (char *) &DATA0_1, 1);
		break;
	case SOUND_MIXER_READ_BASS:
		val = (DATA0_1.bass * 100) / 15;
		val |= val << 8;
		put_user(val, (long *) arg);
		break;
	case SOUND_MIXER_WRITE_LINE1:
		get_user(val, (long *) arg);
		DATA0_ext0.ch1_gain = (((val & 0xff) + 1) * 31) / 100;
		L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0,
			 (char *) &DATA0_ext0, 2);
		break;
	case SOUND_MIXER_READ_LINE1:
		val = (DATA0_ext0.ch1_gain * 100) / 31;
		val |= val << 8;
		put_user(val, (long *) arg);
		break;
	case SOUND_MIXER_WRITE_LINE2:
		get_user(val, (long *) arg);
		DATA0_ext1.ch2_gain = (((val & 0xff) + 1) * 31) / 100;
		L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0,
			 (char *) &DATA0_ext1, 2);
		break;
	case SOUND_MIXER_READ_LINE2:
		val = (DATA0_ext1.ch2_gain * 100) / 31;
		val |= val << 8;
		put_user(val, (long *) arg);
		break;
	case SOUND_MIXER_WRITE_MIC:
		get_user(val, (long *) arg);

		/* Use different registers depending on AGC setting */
		if (DATA0_ext4.AGC_ctrl == 1) {
			/* AGC On, play with MIC sensitivity */
			/* even if 3 bits, value 7 is not used */
			DATA0_ext2.mic_level =
			    (((val & 0xff) + 1) * 6) / 100;
			L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0,
				 (char *) &DATA0_ext2, 2);
		} else {
			/* AGC Off, plain with Input channel 2 amplifer gain */
			val = ((val & 0xff) * 127) / 100;
			DATA0_ext4.ch2_igain_l = val & 3;
			DATA0_ext5.ch2_igain_h = val >> 2;
			L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0,
				 (char *) &DATA0_ext4, 2);
			L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0,
				 (char *) &DATA0_ext5, 2);
		}
		break;
	case SOUND_MIXER_READ_MIC:
		/* Use different registers depending on AGC setting */
		if (DATA0_ext4.AGC_ctrl == 1) {
			val = (DATA0_ext2.mic_level * 100) / 6;
			val |= val << 8;
		} else {
			val =
			    DATA0_ext4.ch2_igain_l +
			    (DATA0_ext5.ch2_igain_h << 2);
			val = (val * 100) / 127;
			val |= val << 8;
		}
		put_user(val, (long *) arg);
		break;

#if 0				/* Experimental.  What those should produce is till not obvious to me. */
	case SOUND_MIXER_WRITE_OGAIN:
		get_user(val, (long *) arg);
		DATA0_ext6.AGC_level = (((val & 0xff) + 1) * 3) / 100;
		L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0,
			 (char *) &DATA0_ext6, 2);
		break;
	case SOUND_MIXER_READ_OGAIN:
		val = (DATA0_ext6.AGC_level * 100) / 3;
		val |= val << 8;
		put_user(val, (long *) arg);
		break;
	case SOUND_MIXER_WRITE_IMIX:
		get_user(val, (long *) arg);
		DATA0_ext2.mixer_mode = val;
		L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0,
			 (char *) &DATA0_ext2, 2);
		break;
	case SOUND_MIXER_READ_IMIX:
		val = DATA0_ext2.mixer_mode;
		put_user(val, (long *) arg);
		break;
#endif

	case SOUND_MIXER_WRITE_RECLEV:
	case SOUND_MIXER_READ_RECLEV:
	default:
		return -EINVAL;
	}
	return 0;
}


static int audio_ioctl(struct inode *inode, struct file *file,
		       uint cmd, ulong arg)
{
	long val;

	/* dispatch based on command */
	switch (cmd) {
	case SNDCTL_DSP_SETFMT:
		get_user(val, (long *) arg);
		if (val & AUDIO_FMT_MASK) {
			audio_fmt = val;
			break;
		} else
			return -EINVAL;

	case SNDCTL_DSP_CHANNELS:
	case SNDCTL_DSP_STEREO:
		get_user(val, (long *) arg);
		if (cmd == SNDCTL_DSP_STEREO)
			val = val ? 2 : 1;
		if (val == 2) {
			audio_channels = val;
			break;
		} else
			return -EINVAL;

	case SOUND_PCM_READ_CHANNELS:
		/* return current number of channels */
		put_user(audio_channels, (long *) arg);
		break;

	case SNDCTL_DSP_SPEED:
		get_user(val, (long *) arg);

		if (machine_is_assabet()) {
		  	if (machine_has_neponset()) {
#ifdef CONFIG_SA1111
			  	/* Note that the 32kHz and 44.1kHz sample
				 * rates are listed in the July 2000 release
				 * of the SA-1111 Developer's Manual.
				 * (Thanks to Chester for pointing this
				 * out! -jd)
				 */
			  	if (val > 0 && val < 9512) {
				  	val = 8000;  /* 8kHz                 */
					SKAUD = 69;
				} else if (val >= 9512 && val < 13512) {
				  	val = 11025; /* 11.025kHz (11.01kHz) */
					SKAUD = 50;
				} else if (val >= 13512 && val < 19025) {
				  	val = 16000; /* 16kHz (16.05kHz)     */
					SKAUD = 34;
				} else if (val >= 19025 && val < 27025){
				  	val = 22050; /* 22.05kHz (22.46kHz)  */
					SKAUD = 24;
				} else if (val >= 27025 && val < 38050){
				  	val = 32000; /* 32kHz (31.2kHz)      */
					SKAUD = 18;
				} else {
				  	val = 44100; /* 44.1kHz (43.2kHz)    */
					SKAUD = 12;
				}

				STATUS_0.system_clk = UDA_STATUS0_SC_256FS;
				L3_write((UDA1341_L3Addr << 2) | 
					 UDA1341_STATUS,
					 (char *) &STATUS_0, 1);
#endif
			} else {
			  	/* FIXME: we need to modify the clock rate for DRAM bank 2 
				 * (our clock source) accordingly.
				 */
			  	val = 44100;
			}

		} else if (machine_is_bitsy()) {
#ifdef CONFIG_SA1100_BITSY
			/* set the external clock generator */
			switch (val) {
			case 32000:
			case 48000:
				/* 00 */
				GPCR =
				    GPIO_BITSY_CLK_SET0 |
				    GPIO_BITSY_CLK_SET1;
				break;
			default:
				val = 44100;
			case 44100:
				/* 01 */
				GPSR = GPIO_BITSY_CLK_SET0;
				GPCR = GPIO_BITSY_CLK_SET1;
				break;
			case 8000:
			case 16000:
				/* 10 */
				GPCR = GPIO_BITSY_CLK_SET0;
				GPSR = GPIO_BITSY_CLK_SET1;
				break;
			case 11025:
			case 22050:
				/* 11 */
				GPSR =
				    GPIO_BITSY_CLK_SET0 |
				    GPIO_BITSY_CLK_SET1;
				break;
			}
			switch (val) {
			case 8000:
				STATUS_0.system_clk = UDA_STATUS0_SC_512FS;
				break;
			case 11025:
				STATUS_0.system_clk = UDA_STATUS0_SC_512FS;
				break;
			case 16000:
				STATUS_0.system_clk = UDA_STATUS0_SC_256FS;
				break;
			case 22050:
				STATUS_0.system_clk = UDA_STATUS0_SC_256FS;
				break;
			case 32000:
				STATUS_0.system_clk = UDA_STATUS0_SC_384FS;
				break;
			case 44100:
				STATUS_0.system_clk = UDA_STATUS0_SC_256FS;
				break;
			case 48000:
				STATUS_0.system_clk = UDA_STATUS0_SC_256FS;
				break;
			}
			L3_write((UDA1341_L3Addr << 2) | UDA1341_STATUS,
				 (char *) &STATUS_0, 1);
#endif
		}

		/* save audio rate */
		audio_rate = val;

		put_user(val, (long *) arg);
		break;

	case SOUND_PCM_READ_RATE:
		put_user(audio_rate, (long *) arg);
		break;

	case SNDCTL_DSP_GETFMTS:
		put_user(AUDIO_FMT_MASK, (long *) arg);
		break;

	case SNDCTL_DSP_GETBLKSIZE:
		put_user(audio_fragsize, (long *) arg);
		break;

	case SNDCTL_DSP_SETFRAGMENT:
		if (output_stream.buffers)
			return -EBUSY;
		get_user(val, (long *) arg);
		audio_fragsize = 1 << (val & 0xFFFF);
		if (audio_fragsize < 16)
			audio_fragsize = 16;
		if (audio_fragsize > 16384)
			audio_fragsize = 16384;
		audio_nbfrags = (val >> 16) & 0x7FFF;
		if (audio_nbfrags < 2)
			audio_nbfrags = 2;
		if (audio_nbfrags * audio_fragsize > 128 * 1024)
			audio_nbfrags = 128 * 1024 / audio_fragsize;
		if (audio_setup_buf(&output_stream))
			return -ENOMEM;
		break;

	case SNDCTL_DSP_SYNC:
		return audio_sync(file);

	case SNDCTL_DSP_GETOSPACE:
		{
			audio_stream_t *s = &output_stream;
			audio_buf_info *inf = (audio_buf_info *) arg;
			int err =
			    verify_area(VERIFY_WRITE, inf, sizeof(*inf));
			int i;
			int frags = 0, bytes = 0;

			if (err)
				return err;
			for (i = 0; i < s->nbfrags; i++) {
				if (atomic_read(&s->buffers[i].sem.count) >
				    0) {
					if (s->buffers[i].size == 0)
						frags++;
					bytes +=
					    s->fragsize -
					    s->buffers[i].size;
				}
			}
			put_user(frags, &inf->fragments);
			put_user(s->nbfrags, &inf->fragstotal);
			put_user(s->fragsize, &inf->fragsize);
			put_user(bytes, &inf->bytes);
		}
		break;

	case SNDCTL_DSP_GETISPACE:
		{
			audio_stream_t *s = &input_stream;
			audio_buf_info *inf = (audio_buf_info *) arg;
			int err =
			    verify_area(VERIFY_WRITE, inf, sizeof(*inf));
			int i;
			int frags = 0, bytes = 0;

			if (err)
				return err;
			for (i = 0; i < s->nbfrags; i++) {
				if (atomic_read(&s->buffers[i].sem.count) >
				    0) {
					if (s->buffers[i].size ==
					    s->fragsize) frags++;
					bytes += s->buffers[i].size;
				}
			}
			put_user(frags, &inf->fragments);
			put_user(s->nbfrags, &inf->fragstotal);
			put_user(s->fragsize, &inf->fragsize);
			put_user(bytes, &inf->bytes);
		}
		break;

	case SNDCTL_DSP_RESET:
		/* ... racy, should be a per fd thing to choose the right stream. */
		//audio_clear_buf(&output_stream);
		return 0;

	case SNDCTL_DSP_POST:
	case SNDCTL_DSP_SUBDIVIDE:
	case SNDCTL_DSP_NONBLOCK:
	case SNDCTL_DSP_GETCAPS:
	case SNDCTL_DSP_GETTRIGGER:
	case SNDCTL_DSP_SETTRIGGER:
	case SNDCTL_DSP_GETIPTR:
	case SNDCTL_DSP_GETOPTR:
	case SNDCTL_DSP_MAPINBUF:
	case SNDCTL_DSP_MAPOUTBUF:
	case SNDCTL_DSP_SETSYNCRO:
	case SNDCTL_DSP_SETDUPLEX:
		return -EINVAL;
	default:
		/* Maybe this is meant for the mixer (As per OSS Docs) */
		return mixer_ioctl(inode, file, cmd, arg);
	}

	return 0;
}


static int audio_open(struct inode *inode, struct file *file)
{
	DPRINTK("audio_open\n");

	if (audio_refcount == 0) {

	  	if (machine_is_assabet() && machine_has_neponset()) {
#ifdef CONFIG_SA1111
			SACR0 |= SACR0_ENB;
#endif
		}

		audio_rate = AUDIO_RATE_DEFAULT;
		audio_fragsize = AUDIO_FRAGSIZE_DEFAULT;
		audio_nbfrags = AUDIO_NBFRAGS_DEFAULT;
		audio_clear_buf(&output_stream);
		audio_clear_buf(&input_stream);
	}
	audio_refcount++;

	MOD_INC_USE_COUNT;
	return 0;
}


static int mixer_open(struct inode *inode, struct file *file)
{
	MOD_INC_USE_COUNT;
	return 0;
}


static int audio_release(struct inode *inode, struct file *file)
{
	DPRINTK("audio_release\n");

	audio_refcount--;
	if (audio_refcount == 0) {
		audio_sync(file);

		sa1100_dma_set_spin(output_stream.dma_ch, 0, 0);

		audio_clear_buf(&output_stream);
		audio_clear_buf(&input_stream);

		if (machine_is_assabet() && machine_has_neponset()) {
#ifdef CONFIG_SA1111
		  	SACR0 &= ~SACR0_ENB;
#endif
#if defined (CONFIG_ASSABET_NEPONSET) && defined (CHROME)
		  	LEDS = 0;
#endif
		}
	}

	MOD_DEC_USE_COUNT;
	return 0;
}


static int mixer_release(struct inode *inode, struct file *file)
{
	MOD_DEC_USE_COUNT;
	return 0;
}


static struct file_operations UDA1341_dsp_fops = {
	llseek:		audio_llseek,
	write:		audio_write,
	read:		audio_read,
	poll:		audio_poll,
	ioctl:		audio_ioctl,
	open:		audio_open,
	release:	audio_release
};

static struct file_operations UDA1341_mixer_fops = {
	ioctl:		mixer_ioctl,
	open:		mixer_open,
	release:	mixer_release
};


/* 
 * TODO: validate the two following function and see if they can be cleanly
 * merged at some point.
 */

static inline void audio_assabet_init(void)
{
#ifdef CONFIG_SA1100_ASSABET
  	if (machine_has_neponset()) {

#ifdef CONFIG_SA1111
        	/* Select I2S audio (instead of AC-Link) */
	  	AUD_CTL = AUD_SEL_1341;
		SKCR &= ~SKCR_SELAC;
		
		/* Enable the I2S clock and L3 bus clock: */
		SKPCR |= (SKPCR_I2SCLKEN | SKPCR_L3CLKEN);
		
		/* Activate and reset the Serial Audio Controller */
		SACR0 |= (SACR0_ENB | SACR0_RST);
		mdelay(5);
		SACR0 &= ~SACR0_RST;
		
		/* For I2S, BIT_CLK is supplied internally. The "SA-1111
		 * Specification Update" mentions that the BCKD bit should
		 * be interpreted as "0 = output". Default clock divider
		 * is 22.05kHz.
		 *
		 * Select I2S, L3 bus. "Recording" and "Replaying" 
		 * (receive and transmit) are enabled.
		 */
		SACR1 = SACR1_L3EN;
#endif

	} else {

	  	/* Setup the uarts */
	  	GAFR |= (GPIO_SSP_TXD | GPIO_SSP_RXD | GPIO_SSP_SCLK |
			 GPIO_SSP_SFRM | GPIO_SSP_CLK);
		GPDR |= (GPIO_SSP_TXD | GPIO_SSP_SCLK | GPIO_SSP_SFRM);
		GPDR &= ~(GPIO_SSP_RXD | GPIO_SSP_CLK);
		PPAR |= PPAR_SPR;
		Ser4SSCR0 = SSCR0_SSE + SSCR0_DataSize(16) + SSCR0_TI;
		Ser4SSCR1 = SSCR1_SClkIactL + SSCR1_SClk1P + SSCR1_ExtClk;
		
		/* Enable the audio power */
		BCR_clear(BCR_STEREO_LB | BCR_QMUTE | BCR_SPK_OFF);
		BCR_set(BCR_CODEC_RST | BCR_AUDIO_ON);
		
		/* 
		 * The assabet board uses the SDRAM clock as the source clock for 
		 * audio. This is supplied to the SA11x0 from the CPLD on pin 19. 
		 * At 206Mhz we need to run the audio clock (SDRAM bank 2) 
		 * at half speed. This clock will scale with core frequency so 
		 * the audio sample rate will also scale. The CPLD on Assabet 
		 * will need to be programmed to match the core frequency.
		 */
		MDREFR |= (MDREFR_K2DB2 | MDREFR_K2RUN);
		/* Make sure EAPD and KAPD are clear to run the clocks at all times. */
		MDREFR &= ~(MDREFR_EAPD | MDREFR_KAPD);
		
		/* Wait for the UDA1341 to wake up */
		mdelay(100);

	}
#endif
}

static inline void audio_assabet_shutdown(void)
{
#ifdef CONFIG_SA1100_ASSABET

	/* disable the audio power */
	BCR_set(BCR_STEREO_LB | BCR_QMUTE | BCR_SPK_OFF);
	BCR_clear(BCR_CODEC_RST | BCR_AUDIO_ON);

#endif
}

static inline void audio_bitsy_init(void)
{
#ifdef CONFIG_SA1100_BITSY
	/* Setup the uarts */
	GAFR &=
	    ~(GPIO_SSP_TXD | GPIO_SSP_RXD | GPIO_SSP_SCLK | GPIO_SSP_SFRM);
	GAFR |= (GPIO_SSP_CLK);
	GPDR &= ~(GPIO_SSP_CLK);
	Ser4SSCR0 = 0;
	Ser4SSCR0 = SSCR0_DataSize(16) + SSCR0_TI;
	Ser4SSCR0 |= SSCR0_SerClkDiv(8);
	Ser4SSCR1 = SSCR1_SClkIactL + SSCR1_SClk1P + SSCR1_ExtClk;
	Ser4SSCR0 |= SSCR0_SSE;

	/* Enable the audio power */
	set_bitsy_egpio(EGPIO_BITSY_CODEC_NRESET | EGPIO_BITSY_AUD_AMP_ON |
			EGPIO_BITSY_AUD_PWR_ON);
	clr_bitsy_egpio(EGPIO_BITSY_QMUTE);
	GPDR |= (GPIO_BITSY_CLK_SET0 | GPIO_BITSY_CLK_SET1);
	/* external clock configured for 44100 samples/sec */
	GPSR = GPIO_BITSY_CLK_SET0;
	GPCR = GPIO_BITSY_CLK_SET1;

	/* Wait for the UDA1341 to wake up */
	mdelay(100);
#endif
}

static inline void audio_bitsy_shutdown(void)
{
#ifdef CONFIG_SA1100_BITSY
	/* disable the audio power */
	set_bitsy_egpio(EGPIO_BITSY_QMUTE);
	clr_bitsy_egpio(EGPIO_BITSY_CODEC_NRESET | EGPIO_BITSY_AUD_AMP_ON |
			EGPIO_BITSY_AUD_PWR_ON);
#endif
}


static int __init audio_init_dma(audio_stream_t * s, char *desc)
{
	int err;

	if(machine_is_assabet() && machine_has_neponset())
	  	err = sa1111_sac_request_dma(&s->dma_ch, desc,
					     (s == &output_stream)?\
					     SA1111_SAC_XMT_CHANNEL:\
					     SA1111_SAC_RCV_CHANNEL);
	else
	  	err = sa1100_request_dma(&s->dma_ch, desc);
	if (err)
		return err;
	if (s == &output_stream) {
		sa1100_dma_set_callback(s->dma_ch,
					audio_dmaout_done_callback);
		sa1100_dma_set_device(s->dma_ch, DMA_Ser4SSPWr);
	} else {
		sa1100_dma_set_callback(s->dma_ch,
					audio_dmain_done_callback);
		sa1100_dma_set_device(s->dma_ch, DMA_Ser4SSPRd);
	}
	return 0;
}


static int audio_clear_dma(audio_stream_t * s)
{
	sa1100_free_dma(s->dma_ch);
	return 0;
}


static void audio_uda1341_reset()
{
	/* Reset the chip */
	STATUS_0.reset = 1;
	L3_write((UDA1341_L3Addr << 2) | UDA1341_STATUS,
		 (char *) &STATUS_0, 1);

	if (machine_is_bitsy()) {
#ifdef CONFIG_BITSY
		clr_bitsy_egpio(EGPIO_BITSY_CODEC_NRESET);
		set_bitsy_egpio(EGPIO_BITSY_CODEC_NRESET);
#endif
	}

	if (machine_has_neponset()) {
	  	STATUS_0.input_fmt = UDA_STATUS0_IF_I2S;
		STATUS_0.system_clk = UDA_STATUS0_SC_256FS;
	}

	STATUS_0.reset = 0;
	L3_write((UDA1341_L3Addr << 2) | UDA1341_STATUS,
		 (char *) &STATUS_0, 1);

}

static int audio_uda1341_pm_callback(struct pm_dev *pm_dev, pm_request_t req, void *data)
{
	switch (req) {
        case PM_SUSPEND: /* enter D1-D3 */
                audio_uda1341_reset();
                /* power down audio */
                if (machine_is_assabet())
                        audio_assabet_shutdown();
                if (machine_is_bitsy())
                        audio_bitsy_shutdown();

                break;
        case PM_RESUME:  /* enter D0 */
                L3_init();
                if (machine_is_assabet())
                        audio_assabet_init();
                if (machine_is_bitsy())
                        audio_bitsy_init();
                audio_uda1341_reset();

                /* Set the mixer values... */
                L3_write((UDA1341_L3Addr << 2) | UDA1341_STATUS, (char *) &STATUS_1, 1);
                L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0, (char *) &DATA0_0, 1);
                L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0, (char *) &DATA0_2, 1);
                L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0, (char *) &DATA0_ext2, 2);
                L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0, (char *) &DATA0_ext4, 2);
                L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0, (char *) &DATA0_ext6, 2);

                break;
	}
	return 0;
}

int __init audio_uda1341_init(void)
{

	if (!(machine_is_assabet() || machine_is_bitsy()))
		return -ENODEV;

	/* Acquire and initialize DMA */
	if (audio_init_dma(&output_stream, "UDA1341 out") ||
	    audio_init_dma(&input_stream, "UDA1341 in")) {
		audio_clear_dma(&output_stream);
		audio_clear_dma(&input_stream);
		return -EBUSY;
	}

	L3_init();

	if (machine_is_assabet())
		audio_assabet_init();
	if (machine_is_bitsy())
		audio_bitsy_init();
	audio_uda1341_reset();

	/* Set some default mixer values... */
	STATUS_1.DAC_gain = 1;
	STATUS_1.ADC_gain = 1;
	L3_write((UDA1341_L3Addr << 2) | UDA1341_STATUS,
		 (char *) &STATUS_1, 1);
	DATA0_0.volume = 15;
	L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0, (char *) &DATA0_0,
		 1);
	DATA0_2.mode = 3;
	L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0, (char *) &DATA0_2,
		 1);
	DATA0_ext2.mixer_mode = 2;
	DATA0_ext2.mic_level = 4;
	L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0,
		 (char *) &DATA0_ext2, 2);
	DATA0_ext4.AGC_ctrl = 1;
	L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0,
		 (char *) &DATA0_ext4, 2);
	DATA0_ext6.AGC_level = 3;
	L3_write((UDA1341_L3Addr << 2) | UDA1341_DATA0,
		 (char *) &DATA0_ext6, 2);

	/* register devices */
	audio_dev_dsp = register_sound_dsp(&UDA1341_dsp_fops, -1);
	audio_dev_mixer = register_sound_mixer(&UDA1341_mixer_fops, -1);

#ifdef CONFIG_PM
        audio_pm_dev = pm_register(PM_SYS_DEV, 0, audio_uda1341_pm_callback);
#endif

	printk(AUDIO_NAME_VERBOSE "%s initialized\n",
	       machine_has_neponset() ? " (SA-1111 L3 Control Bus)" : "" );

	return 0;
}

module_init(audio_uda1341_init);


void __exit audio_uda1341_exit(void)
{
	/* unregister driver and IRQ */
	unregister_sound_dsp(audio_dev_dsp);
	unregister_sound_mixer(audio_dev_mixer);
	audio_clear_dma(&output_stream);
	audio_clear_dma(&input_stream);
	printk(AUDIO_NAME_VERBOSE " unloaded\n");
}

module_exit(audio_uda1341_exit);
